from gaphor import UML
from gaphor.diagram.general import CommentItem
from gaphor.diagram.classes import AssociationItem, ClassItem
from gaphor.services.copyservice import CopyService
from gaphor.tests.testcase import TestCase
from gaphor.storage.verify import orphan_references


class CopyServiceTestCase(TestCase):

    services = TestCase.services + [
        "main_window",
        "properties",
        "undo_manager",
        "export_menu",
        "tools_menu",
    ]

    def setUp(self):
        super().setUp()
        self.service = CopyService(
            self.get_service("event_manager"),
            self.get_service("element_factory"),
            self.get_service("main_window"),
        )

    def test_copy(self):
        service = self.service

        ef = self.element_factory
        diagram = ef.create(UML.Diagram)
        ci = diagram.create(CommentItem, subject=ef.create(UML.Comment))

        service.copy([ci])
        assert diagram.canvas.get_all_items() == [ci]

        service.paste(diagram)

        assert len(diagram.canvas.get_all_items()) == 2, diagram.canvas.get_all_items()

    def _skip_test_copy_paste_undo(self):
        """
        Test if copied data is undoable.
        """

        service = self.service

        # Setting the stage:
        ci1 = self.create(ClassItem, UML.Class)
        ci2 = self.create(ClassItem, UML.Class)
        a = self.create(AssociationItem)

        self.connect(a, a.head, ci1)
        self.connect(a, a.tail, ci2)

        assert a.subject
        assert a.head_end.subject
        assert a.tail_end.subject

        # The act: copy and paste, perform undo afterwards
        service.copy([ci1, ci2, a])

        service.paste(self.diagram)

        all_items = list(self.diagram.canvas.get_all_items())

        assert 6 == len(all_items)
        assert not orphan_references(self.element_factory)

        assert all_items[0].subject is all_items[3].subject
        assert all_items[1].subject is all_items[4].subject
        assert all_items[2].subject is all_items[5].subject

        undo_manager = self.get_service("undo_manager")

        undo_manager.undo_transaction()

        assert 3 == len(self.diagram.canvas.get_all_items())
        assert not orphan_references(self.element_factory)



from gi.repository import Gtk

from gaphor import UML
from gaphor.core import _, transactional
from gaphor.diagram.propertypages import PropertyPages, PropertyPageBase


@PropertyPages.register(UML.Comment)
class CommentItemPropertyPage(PropertyPageBase):
    """Property page for Comments."""

    order = 0

    def __init__(self, subject):
        self.subject = subject
        self.watcher = subject and subject.watcher()

    def construct(self):
        subject = self.subject
        page = Gtk.Box.new(Gtk.Orientation.VERTICAL, 0)

        if not subject:
            return page

        label = Gtk.Label(label=_("Comment"))
        label.set_justify(Gtk.Justification.LEFT)
        page.pack_start(label, False, True, 0)

        buffer = Gtk.TextBuffer()
        if subject.body:
            buffer.set_text(subject.body)
        text_view = Gtk.TextView()
        text_view.set_buffer(buffer)
        text_view.show()
        text_view.set_size_request(-1, 100)
        page.pack_start(text_view, True, True, 0)

        changed_id = buffer.connect("changed", self._on_body_change)

        def handler(event):
            if not text_view.props.has_focus:
                buffer.handler_block(changed_id)
                buffer.set_text(event.new_value)
                buffer.handler_unblock(changed_id)

        self.watcher.watch("body", handler).subscribe_all()
        text_view.connect("destroy", self.watcher.unsubscribe_all)

        return page

    @transactional
    def _on_body_change(self, buffer):
        self.subject.body = buffer.get_text(
            buffer.get_start_iter(), buffer.get_end_iter(), False
        )



"""The properties module allows Gaphor properties to be saved to the local
file system.  These are things like preferences."""

from typing import Dict
import os
import pprint
import sys
import ast

from gaphas.decorators import AsyncIO
from gaphor.misc import get_config_dir


from gaphor.abc import Service


class PropertyChanged:
    """
    This event is triggered any time a property is changed.  This event
    holds the property key, the current value, and the new value.
    """

    def __init__(self, key, old_value, new_value):
        self.key = key
        self.old_value = old_value
        self.new_value = new_value


_no_default = object()


class Properties(Service):
    """The Properties class holds a collection of application wide properties.

    Properties are persisted to the local file system."""

    def __init__(self, event_manager, backend=None):
        """Constructor.  Initialize the Gaphor application object, the
        dictionary for storing properties in memory, and the storage backend.
        This defaults to FileBackend"""
        self.event_manager = event_manager
        self._resources: Dict[str, object] = {}
        self._backend = backend or FileBackend()
        self._backend.load(self._resources)

    def shutdown(self):
        """Shutdown the properties service.  This will ensure that all
        properties are saved."""

        self._backend.save(self._resources)

    def __call__(self, key, default=_no_default):
        """Retrieve the specified property.  If the property doesn't exist,
        the default parameter is returned.  This defaults to _no_default."""

        return self.get(key, default)

    def save(self):
        """Save all properties by calling save() on the properties storage
        backend."""

        self._backend.save(self._resources)

    def _items(self):
        """Return an iterator for all stored properties."""

        return iter(self._resources.items())

    def dump(self, stream=sys.stdout):
        """
        TODO: define resources that are persistent (have to be saved
        and loaded.
        """
        pprint.pprint(list(self._resources.items()), stream)

    def get(self, key: str, default=_no_default):
        """Locate a property.

        Resource should be the class of the resource to look for or a string. In
        case of a string the resource will be looked up in the GConf
        configuration."""

        try:
            return self._resources[key]
        except KeyError:
            if default is _no_default:
                raise KeyError(f'No resource with name "{key}"')

            self.set(key, default)
            return default

    def set(self, key, value):
        """Set a property to a specific value.

        No smart things are done with classes and class names (like the
        resource() method does)."""

        resources = self._resources
        old_value = resources.get(key)

        if value != old_value:
            resources[key] = value
            self.event_manager.handle(PropertyChanged(key, old_value, value))
            self._backend.update(resources, key, value)


class FileBackend:
    """Resource backend that stores data to a resource file
    ($HOME/.gaphor/resource)."""

    RESOURCE_FILE = "resources"

    def __init__(self, datadir=get_config_dir()):
        """Constructor.  Initialize the directory used for storing
        properties."""

        self.datadir = datadir

    def get_filename(self, create=False):
        """Return the current file used to store Gaphor properties.  If the
        created parameter is set to True, the file is created if it doesn't
        exist.  This defaults to False."""

        datadir = self.datadir

        if create and not os.path.exists(datadir):
            os.mkdir(datadir)

        return os.path.join(datadir, self.RESOURCE_FILE)

    def load(self, resource):
        """Load resources from a file. Resources are saved like you do with
        a dict()."""

        filename = self.get_filename()

        if os.path.exists(filename) and os.path.isfile(filename):

            with open(filename) as ifile:
                data = ifile.read()

            for key, value in ast.literal_eval(data).items():
                resource[key] = value

    def save(self, resource):
        """Save persist resources from the resources dictionary.
        @resource is the Resource instance
        @persistent is a list of persistent resource names.
        """

        filename = self.get_filename(create=True)

        with open(filename, "w") as ofile:
            pprint.pprint(resource, ofile)

    @AsyncIO(single=True, timeout=500)
    def update(self, resource, key, value):
        """Update the properties file with any changes in the background."""

        self.save(resource)



"""
Package diagram item.
"""

from gaphor import UML
from gaphor.UML.modelfactory import stereotypes_str
from gaphor.diagram.presentation import ElementPresentation, Named, from_package_str
from gaphor.diagram.shapes import Box, EditableText, Text
from gaphor.diagram.text import FontWeight
from gaphor.diagram.support import represents


@represents(UML.Package)
@represents(UML.Profile)
class PackageItem(ElementPresentation, Named):
    def __init__(self, id=None, model=None):
        super().__init__(id, model)

        self.shape = Box(
            Text(
                text=lambda: stereotypes_str(
                    self.subject,
                    isinstance(self.subject, UML.Profile) and ("profile",) or (),
                ),
                style={"min-width": 0, "min-height": 0},
            ),
            EditableText(
                text=lambda: self.subject and self.subject.name or "",
                style={"font-weight": FontWeight.BOLD},
            ),
            Text(
                text=lambda: from_package_str(self),
                style={"font": "sans 8", "min-width": 0, "min-height": 0},
            ),
            style={"min-width": 50, "min-height": 70, "padding": (25, 10, 5, 10)},
            draw=draw_package,
        )

        self.watch("subject[NamedElement].name")
        self.watch("subject[NamedElement].namespace.name")
        self.watch("subject.appliedStereotype.classifier.name")


def draw_package(box, context, bounding_box):
    cr = context.cairo
    o = 0.0
    h = bounding_box.height
    w = bounding_box.width
    x = 50
    y = 20
    cr.move_to(x, y)
    cr.line_to(x, o)
    cr.line_to(o, o)
    cr.line_to(o, h)
    cr.line_to(w, h)
    cr.line_to(w, y)
    cr.line_to(o, y)
    cr.stroke()



from gaphor.diagram.interactions.interaction import InteractionItem
from gaphor.diagram.interactions.message import MessageItem
from gaphor.diagram.interactions.lifeline import LifelineItem


def _load():
    from gaphor.diagram.interactions import (
        messageconnect,
        interactionsgrouping,
        interactionspropertypages,
    )


_load()



from gaphor.tests import TestCase
from gaphor.diagram.profiles.metaclasspropertypage import MetaclassNamePropertyPage
from gaphor import UML
from gi.repository import Gtk


class MetaclassPropertyPageTest(TestCase):
    def test_name_input_field_for_normal_class(self):
        class_ = self.element_factory.create(UML.Class)

        class_.name = "Class"
        editor = MetaclassNamePropertyPage(class_)
        page = editor.construct()
        assert page
        entry = page.get_children()[0].get_children()[1]
        assert Gtk.Entry is type(entry)

        assert "Class" == entry.get_text()

        class_.name = "Blah"
        assert "Blah" == entry.get_text()

    def test_name_selection_for_metaclass(self):
        metaclass = self.element_factory.create(UML.Class)
        metaclass.name = "Class"
        stereotype = self.element_factory.create(UML.Stereotype)
        stereotype.name = "NewStereotype"
        UML.model.create_extension(metaclass, stereotype)

        editor = MetaclassNamePropertyPage(metaclass)
        page = editor.construct()
        assert page
        combo = page.get_children()[0].get_children()[1]
        assert Gtk.ComboBox is type(combo)

        assert "Class" == combo.get_child().get_text()

        metaclass.name = "Blah"
        assert "Blah" == combo.get_child().get_text()



from gi.repository import Gdk, Gtk
from gaphor.plugins.console.console import main, GTKInterpreterConsole, Help


class KeyEvent:
    def __init__(self, keyval, state=0):
        self.keyval = keyval
        self.state = state


def test_console_opening():
    main(main_loop=False)
    Gtk.main_iteration()


def test_help():
    help = Help()

    assert help() == "Usage: help(object)"


def test_help_on_object(capsys):
    help = Help()

    help(str)
    captured = capsys.readouterr()

    assert "|" in captured.out


def console_text(console):
    buffer = console.text.get_buffer()
    return buffer.get_text(buffer.get_start_iter(), buffer.get_end_iter(), False)


def test_history():
    console = GTKInterpreterConsole()
    console.history.append("my_history()")

    console.key_pressed(console, KeyEvent(Gdk.keyval_from_name("Up")))
    text = console_text(console)

    assert ">>> my_history()" in text


def test_deep_history():
    console = GTKInterpreterConsole()
    console.history.append("deepest()")
    console.history.append("deeper()")
    console.history.append("deep()")

    console.key_pressed(console, KeyEvent(Gdk.keyval_from_name("Up")))
    console.key_pressed(console, KeyEvent(Gdk.keyval_from_name("Up")))
    console.key_pressed(console, KeyEvent(Gdk.keyval_from_name("Up")))
    console.key_pressed(console, KeyEvent(Gdk.keyval_from_name("Down")))
    text = console_text(console)

    assert ">>> deeper()" in text


def test_run_line():
    console = GTKInterpreterConsole()

    console.buffer.append("help")
    console.key_pressed(console, KeyEvent(Gdk.keyval_from_name("Return")))

    text = console_text(console)

    assert "Usage: help(object)" in text



"""
All Item's defined in the diagram package. This module
makes it easier to load a diagram item.
"""

from gaphor.diagram.actions import *
from gaphor.diagram.classes import *
from gaphor.diagram.components import *
from gaphor.diagram.general import *
from gaphor.diagram.interactions import *
from gaphor.diagram.profiles import *
from gaphor.diagram.states import *
from gaphor.diagram.usecases import *



from gaphor.tests import TestCase
from gaphor import UML
from gaphor.diagram.actions import ObjectNodeItem
from gaphor.diagram.editors import Editor


class EditorTestCase(TestCase):
    def setUp(self):
        super().setUp()

    def test_objectnode_editor(self):
        node = self.create(ObjectNodeItem, UML.ObjectNode)
        self.diagram.canvas.update_now()

        adapter = Editor(node)
        print(node, adapter)
        assert adapter.is_editable(10, 10)
        # assert not adapter.edit_tag

        # assert adapter.is_editable(*node.tag_bounds[:2])
        # assert adapter.edit_tag



"""
Test transitions.
"""

from gaphor import UML
from gaphor.diagram.states.transition import TransitionItem
from gaphor.tests.testcase import TestCase


class TransitionTestCase(TestCase):
    """
    Test the working of transitions
    """

    def test_transition_guard(self):
        """Test events of transition.guard.
        """
        item = self.create(TransitionItem, UML.Transition)
        guard = item.shape_middle
        assert guard.text() == ""

        c = self.element_factory.create(UML.Constraint)
        c.specification = "blah"
        assert guard.text() == ""

        item.subject.guard = c
        assert item.subject.guard is c
        assert guard.text() == "blah", guard.text()

        del c.specification
        assert guard.text() == "", guard.text()

        c.specification = "foo"
        assert guard.text() == "foo", item._guard.text()



"""
Use case inclusion relationship.
"""

from gaphor import UML
from gaphor.UML.modelfactory import stereotypes_str
from gaphor.diagram.presentation import LinePresentation
from gaphor.diagram.shapes import Text, draw_arrow_head
from gaphor.diagram.support import represents


@represents(UML.Include)
class IncludeItem(LinePresentation):
    """
    Use case inclusion relationship.
    """

    def __init__(self, id=None, model=None):
        super().__init__(id, model, style={"dash-style": (7.0, 5.0)})

        self.shape_middle = Text(
            text=lambda: stereotypes_str(self.subject, ("include",)),
            style={"min-width": 0, "min-height": 0},
        )
        self.watch("subject.appliedStereotype.classifier.name")
        self.draw_head = draw_arrow_head



"""
Editors.
"""
import abc
from functools import singledispatch

from gaphor import UML
from gaphor.diagram.presentation import Named, Classified


@singledispatch
def Editor(obj):
    pass


@singledispatch
def editable(el):
    """
    Return editable part of UML element.

    It returns element itself by default.
    """
    return el


@editable.register(UML.Slot)
def editable_slot(el):
    """
    Return editable part of a slot.
    """
    return el.value


class AbstractEditor(metaclass=abc.ABCMeta):
    """
    Provide an interface for editing text with the TextEditTool.
    """

    @abc.abstractmethod
    def is_editable(self, x, y):
        """
        Is this item editable in it's current state.
        x, y represent the cursors (x, y) position.
        (this method should be called before get_text() is called.
        """

    @abc.abstractmethod
    def get_text(self):
        """
        Get the text to be updated
        """

    @abc.abstractmethod
    def update_text(self, text):
        """
        Update with the new text.
        """

    @abc.abstractmethod
    def key_pressed(self, pos, key):
        """
        Called every time a key is pressed. Allows for 'Enter' as escape
        character in single line editing.
        """


@Editor.register(Named)
class NamedItemEditor(AbstractEditor):
    """Text edit support for Named items."""

    def __init__(self, item):
        self._item = item

    def is_editable(self, x, y):
        return True

    def get_text(self):
        s = self._item.subject
        return s.name if s else ""

    def update_text(self, text):
        if self._item.subject:
            self._item.subject.name = text
        self._item.request_update()

    def key_pressed(self, pos, key):
        pass


@Editor.register(Classified)
class ClassifiedItemEditor(AbstractEditor):
    """Text editor support for compartment items."""

    def __init__(self, item):
        self._item = item
        self._edit = None

    def is_editable(self, x, y):
        """
        Find out what's located at point (x, y), is it in the
        name part or is it text in some compartment
        """
        self._edit = hasattr(self._item, "item_at") and self._item.item_at(x, y)
        return bool(self._edit)

    def get_text(self):
        return UML.format(editable(self._edit.subject))

    def update_text(self, text):
        UML.parse(editable(self._edit.subject), text)

    def key_pressed(self, pos, key):
        pass



import itertools


# See https://www.omg.org/spec/UML/2.5/PDF, section 12.4.1.5, page 271
def extension_metaclass(self):
    """
    References the Class that is extended through an Extension. The
    property is derived from the type of the memberEnd that is not the
    ownedEnd.
    """
    ownedEnd = self.ownedEnd
    metaend = [e for e in self.memberEnd if e is not ownedEnd]
    if metaend:
        return metaend[0].type


def property_opposite(self):
    """
    In the case where the property is one navigable end of a binary
    association with both ends navigable, this gives the other end.

    For Gaphor the property on the other end is returned regardless the
    navigability.
    """
    if self.association is not None and len(self.association.memberEnd) == 2:
        return (
            self.association.memberEnd[0] is self
            and self.association.memberEnd[1]
            or self.association.memberEnd[0]
        )
    return None


def property_navigability(self):
    """
    Get navigability of an association end.
    If no association is related to the property, then unknown navigability
    is assumed.
    """
    from gaphor.UML.uml2 import Class, Interface

    assoc = self.association
    if not assoc or not self.opposite:
        return None  # assume unknown
    owner = self.opposite.type
    if owner and (
        (type(self.type) in (Class, Interface) and self in owner.ownedAttribute)
        or self in assoc.navigableOwnedEnd
    ):
        return True
    elif self in assoc.ownedEnd:
        return None
    else:
        return False


def _pr_interface_deps(classifier, dep_type):
    """
    Return all interfaces, which are connected to a classifier with given
    dependency type.
    """
    from gaphor.UML.uml2 import Interface

    return (
        dep.supplier[0]
        for dep in classifier.clientDependency
        if dep.isKindOf(dep_type) and dep.supplier[0].isKindOf(Interface)
    )


def _pr_rc_interface_deps(component, dep_type):
    """
    Return all interfaces, which are connected to realizing classifiers of
    specified component. Returned interfaces are connected to realizing
    classifiers with given dependency type.

    Generator of generators is returned. Do not forget to flat it later.
    """
    return (
        _pr_interface_deps(r.realizingClassifier, dep_type)
        for r in component.realization
    )


def component_provided(self):
    """
    Interfaces provided to component environment.
    """
    from gaphor.UML.uml2 import Implementation, Realization

    implementations = (
        impl.contract[0]
        for impl in self.implementation
        if impl.isKindOf(Implementation)
    )
    realizations = _pr_interface_deps(self, Realization)

    # realizing classifiers realizations
    # this generator of generators, so flatten it later
    rc_realizations = _pr_rc_interface_deps(self, Realization)

    return tuple(set(itertools.chain(implementations, realizations, *rc_realizations)))


def component_required(self):
    """
    Interfaces required by component.
    """
    from gaphor.UML.uml2 import Usage

    usages = _pr_interface_deps(self, Usage)

    # realizing classifiers usages
    # this generator of generators, so flatten it later
    rc_usages = _pr_rc_interface_deps(self, Usage)

    return tuple(set(itertools.chain(usages, *rc_usages)))


def message_messageKind(self):
    """
    MessageKind
    """
    kind = "unknown"
    if self.sendEvent:
        kind = "lost"
        if self.receiveEvent:
            kind = "complete"
    elif self.receiveEvent:
        kind = "found"
    return kind


def namedelement_qualifiedname(self):
    """
    Returns the qualified name of the element as a tuple
    """
    if self.namespace:
        return self.namespace.qualifiedName + (self.name,)
    else:
        return (self.name,)



from gaphas.canvas import Context, instant_cairo_context
import gaphor.UML as UML
from gaphor.tests.testcase import TestCase
from gaphor.diagram.actions.flow import FlowItem


class FlowTestCase(TestCase):
    def test_flow(self):
        self.create(FlowItem, UML.ControlFlow)

    def test_name(self):
        """
        Test updating of flow name text.
        """
        flow = self.create(FlowItem, UML.ControlFlow)
        name = flow.shape_tail.children[1]

        flow.subject.name = "Blah"

        assert "Blah" == name.text()

        flow.subject = None

        assert "" == name.text()

    def test_guard_text_update(self):
        flow = self.create(FlowItem, UML.ControlFlow)
        guard = flow.shape_middle

        assert "" == guard.text()

        flow.subject.guard = "GuardMe"
        assert "GuardMe" == guard.text()

        flow.subject = None
        assert "" == guard.text()

    def test_draw(self):
        flow = self.create(FlowItem, UML.ControlFlow)
        context = Context(
            cairo=instant_cairo_context(), hovered=True, focused=True, selected=True
        )

        flow.draw(context)



from gaphor import UML
from gaphor.diagram.editors import Editor, AbstractEditor
from gaphor.diagram.classes.association import AssociationItem


@Editor.register(AssociationItem)
class AssociationItemEditor(AbstractEditor):
    def __init__(self, item):
        self._item = item
        self._edit: AssociationItem = None

    def is_editable(self, x, y):
        """Find out what's located at point (x, y), is it in the
        name part or is it text in some compartment
        """
        item = self._item
        if not item.subject:
            return False
        if item.head_end.point((x, y)) <= 0:
            self._edit = item.head_end
        elif item.tail_end.point((x, y)) <= 0:
            self._edit = item.tail_end
        else:
            self._edit = item
        return True

    def get_text(self):
        if self._edit is self._item:
            return self._edit.subject.name
        return UML.format(
            self._edit.subject,
            visibility=True,
            is_derived=True,
            type=True,
            multiplicity=True,
            default=True,
        )

    def update_text(self, text):
        UML.parse(self._edit.subject, text)

    def key_pressed(self, pos, key):
        pass



import pytest
from gaphor.UML.element import Element
from gaphor.UML.event import AssociationUpdated
from gaphor.UML.properties import *
from gaphor.application import Application
from gaphor.core import event_handler


def test_association_1_x():
    #
    # 1:-
    #
    class A(Element):
        pass

    class B(Element):
        pass

    A.one = association("one", B, 0, 1, opposite="two")
    B.two = association("two", A, 0, 1)
    a = A()
    b = B()
    a.one = b
    assert a.one is b
    assert b.two is a

    a.one = None
    assert a.one is None
    assert b.two is None

    a.one = b
    assert a.one is b
    assert b.two is a

    del a.one
    assert a.one is None
    assert b.two is None


def test_association_n_x():
    #
    # n:-
    #
    class A(Element):
        pass

    class B(Element):
        pass

    class C(Element):
        pass

    A.one = association("one", B, 0, "*", opposite="two")
    B.two = association("two", A, 0, 1)

    a = A()
    b = B()
    a.one = b
    assert b in a.one
    assert b.two is a


def test_association_1_1():
    #
    # 1:1
    #
    class A(Element):
        pass

    class B(Element):
        pass

    class C(Element):
        pass

    A.one = association("one", B, 0, 1, opposite="two")
    B.two = association("two", A, 0, 1, opposite="one")

    a = A()
    b = B()
    a.one = b
    a.one = b

    assert a.one is b
    assert b.two is a

    a.one = B()
    assert a.one is not b
    assert b.two is None

    c = C()
    try:
        a.one = c
    except Exception as e:
        pass  # ok
    else:
        assert a.one is not c

    del a.one
    assert a.one is None
    assert b.two is None


def test_association_1_n():
    #
    # 1:n
    #
    class A(Element):
        pass

    class B(Element):
        pass

    class C(Element):
        pass

    A.one = association("one", B, lower=0, upper=1, opposite="two")
    B.two = association("two", A, lower=0, upper="*", opposite="one")

    a1 = A()
    a2 = A()
    b1 = B()
    b2 = B()

    b1.two = a1
    assert len(b1.two) == 1, "len(b1.two) == %d" % len(b1.two)
    assert a1 in b1.two
    assert a1.one is b1, f"{a1.one}/{b1}"
    b1.two = a1
    b1.two = a1
    assert len(b1.two) == 1, "len(b1.two) == %d" % len(b1.two)
    assert a1 in b1.two
    assert a1.one is b1, f"{a1.one}/{b1}"

    b1.two = a2
    assert a1 in b1.two
    assert a2 in b1.two
    assert a1.one is b1
    assert a2.one is b1

    try:
        del b1.two
    except Exception:
        pass  # ok
    else:
        assert b1.two != []

    assert a1 in b1.two
    assert a2 in b1.two
    assert a1.one is b1
    assert a2.one is b1

    b1.two.remove(a1)

    assert len(b1.two) == 1
    assert a1 not in b1.two
    assert a2 in b1.two
    assert a1.one is None
    assert a2.one is b1

    a2.one = b2

    assert len(b1.two) == 0
    assert len(b2.two) == 1
    assert a2 in b2.two
    assert a1.one is None
    assert a2.one is b2

    try:
        del b1.two[a1]
    except ValueError:
        pass  # ok
    else:
        assert 0, "should not be removed"


def test_association_n_n():
    #
    # n:n
    #
    class A(Element):
        pass

    class B(Element):
        pass

    class C(Element):
        pass

    A.one = association("one", B, 0, "*", opposite="two")
    B.two = association("two", A, 0, "*", opposite="one")

    a1 = A()
    a2 = A()
    b1 = B()
    b2 = B()

    a1.one = b1
    assert b1 in a1.one
    assert a1 in b1.two
    assert not a2.one
    assert not b2.two

    a1.one = b2
    assert b1 in a1.one
    assert b2 in a1.one
    assert a1 in b1.two
    assert a1 in b2.two
    assert not a2.one

    a2.one = b1
    assert len(a1.one) == 2
    assert len(a2.one) == 1
    assert len(b1.two) == 2
    assert len(b2.two) == 1
    assert b1 in a1.one
    assert b2 in a1.one
    assert a1 in b1.two
    assert a1 in b2.two
    assert b1 in a2.one
    assert a2 in b1.two

    del a1.one[b1]
    assert len(a1.one) == 1
    assert len(a2.one) == 1
    assert len(b1.two) == 1
    assert len(b2.two) == 1
    assert b1 not in a1.one
    assert b2 in a1.one
    assert a1 not in b1.two
    assert a1 in b2.two
    assert b1 in a2.one
    assert a2 in b1.two


def test_association_swap():
    class A(Element):
        pass

    class B(Element):
        pass

    class C(Element):
        pass

    A.one = association("one", B, 0, "*")

    a = A()
    b1 = B()
    b2 = B()

    a.one = b1
    a.one = b2
    assert a.one.size() == 2
    assert a.one[0] is b1
    assert a.one[1] is b2

    events = []

    @event_handler(AssociationUpdated)
    def handler(event, events=events):
        events.append(event)

    #        Application.register_handler(handler)
    #        try:
    a.one.swap(b1, b2)
    #            assert len(events) == 1
    #            assert events[0].property is A.one
    #            assert events[0].element is a
    #        finally:
    #            Application.unregister_handler(handler)

    assert a.one.size() == 2
    assert a.one[0] is b2
    assert a.one[1] is b1


def test_association_unlink_1():
    #
    # unlink
    #
    class A(Element):
        pass

    class B(Element):
        pass

    class C(Element):
        pass

    A.one = association("one", B, 0, "*")

    a1 = A()
    a2 = A()
    b1 = B()
    b2 = B()

    a1.one = b1
    a1.one = b2
    assert b1 in a1.one
    assert b2 in a1.one

    a2.one = b1

    # remove b1 from all elements connected to b1
    # also the signal should be removed
    b1.unlink()

    assert b1 not in a1.one
    assert b2 in a1.one


def test_association_unlink_2():
    #
    # unlink
    #
    class A(Element):
        pass

    class B(Element):
        pass

    class C(Element):
        pass

    A.one = association("one", B, 0, "*", opposite="two")
    B.two = association("two", A, 0, "*")

    a1 = A()
    a2 = A()
    b1 = B()
    b2 = B()

    a1.one = b1
    a1.one = b2
    assert b1 in a1.one
    assert b2 in a1.one
    assert a1 in b1.two
    assert a1 in b2.two

    a2.one = b1

    # remove b1 from all elements connected to b1
    # also the signal should be removed
    b1.unlink()

    assert b1 not in a1.one
    assert b2 in a1.one
    assert a1 not in b1.two
    assert a1 in b2.two


def test_attributes():
    class A(Element):
        pass

    A.a = attribute("a", bytes, "default")

    a = A()
    assert a.a == "default", a.a
    a.a = "bar"
    assert a.a == "bar", a.a
    del a.a
    assert a.a == "default"
    try:
        a.a = 1
    except AttributeError:
        pass  # ok
    else:
        assert 0, "should not set integer"


def test_enumerations():
    class A(Element):
        pass

    A.a = enumeration("a", ("one", "two", "three"), "one")
    a = A()
    assert a.a == "one"
    a.a = "two"
    assert a.a == "two"
    a.a = "three"
    assert a.a == "three"
    try:
        a.a = "four"
    except AttributeError:
        assert a.a == "three"
    else:
        assert 0, "a.a could not be four"
    del a.a
    assert a.a == "one"


@pytest.mark.skip
def test_notify():
    class A(Element):
        notified = None

        def notify(self, name, pspec):
            self.notified = name

    A.assoc = association("assoc", A)
    A.attr = attribute("attr", bytes, "default")
    A.enum = enumeration("enum", ("one", "two"), "one")

    a = A()
    assert a.notified == None
    a.assoc = A()
    assert a.notified == "assoc", a.notified
    a.attr = "newval"
    assert a.notified == "attr", a.notified
    a.enum = "two"
    assert a.notified == "enum", a.notified
    a.notified = None
    a.enum = "two"  # should not notify since value hasn't changed.
    assert a.notified == None


def test_derivedunion():
    class A(Element):
        pass

    A.a = association("a", A)
    A.b = association("b", A, 0, 1)
    A.u = derivedunion("u", object, 0, "*", A.a, A.b)

    a = A()
    assert len(a.a) == 0, f"a.a = {a.a}"
    assert len(a.u) == 0, f"a.u = {a.u}"
    a.a = b = A()
    a.a = c = A()
    assert len(a.a) == 2, f"a.a = {a.a}"
    assert b in a.a
    assert c in a.a
    assert len(a.u) == 2, f"a.u = {a.u}"
    assert b in a.u
    assert c in a.u

    a.b = d = A()
    assert len(a.a) == 2, f"a.a = {a.a}"
    assert b in a.a
    assert c in a.a
    assert d == a.b
    assert len(a.u) == 3, f"a.u = {a.u}"
    assert b in a.u
    assert c in a.u
    assert d in a.u


@pytest.mark.skip
def test_derivedunion_notify():
    class A(Element):
        pass

    class E(Element):
        notified = False

        def notify(self, name, pspec):
            if name == "u":
                self.notified = True

    E.a = association("a", A)
    E.u = derivedunion("u", A, 0, "*", E.a)

    e = E()
    assert e.notified == False
    e.a = A()
    assert e.notified == True


def test_derivedunion_listmixins():
    class A(Element):
        pass

    A.a = association("a", A)
    A.b = association("b", A)
    A.u = derivedunion("u", A, 0, "*", A.a, A.b)
    A.name = attribute("name", str, "default")

    a = A()
    a.a = A()
    a.a = A()
    a.b = A()
    a.a[0].name = "foo"
    a.a[1].name = "bar"
    a.b[0].name = "baz"

    assert list(a.a[:].name) == ["foo", "bar"]
    assert sorted(list(a.u[:].name)) == ["bar", "baz", "foo"]


def test_composite():
    class A(Element):
        is_unlinked = False

        def unlink(self):
            self.is_unlinked = True
            Element.unlink(self)

    A.comp = association("comp", A, composite=True, opposite="other")
    A.other = association("other", A, composite=False, opposite="comp")

    a = A()
    a.name = "a"
    b = A()
    b.name = "b"
    a.comp = b
    assert b in a.comp
    assert a in b.other

    a.unlink()
    assert a.is_unlinked
    assert b.is_unlinked


@pytest.mark.skip
def test_derivedunion_events():
    from gaphor.UML.event import DerivedAdded, DerivedDeleted

    class A(Element):
        is_unlinked = False

        def unlink(self):
            self.is_unlinked = True
            Element.unlink(self)

    A.a1 = association("a1", A, upper=1)
    A.a2 = association("a2", A, upper=1)
    A.b1 = association("b1", A, upper="*")
    A.b2 = association("b2", A, upper="*")
    A.b3 = association("b3", A, upper=1)

    A.derived_a = derivedunion("derived_a", object, 0, 1, A.a1, A.a2)
    A.derived_b = derivedunion("derived_b", object, 0, "*", A.b1, A.b2, A.b3)

    events = []

    @event_handler(AssociationUpdated)
    def handler(event, events=events):
        events.append(event)

    Application.register_handler(handler)
    try:
        a = A()
        a.a1 = A()
        assert len(events) == 2
        assert events[0].property is A.derived_a
        assert events[1].property is A.a1
        assert a.derived_a is a.a1
        a.a1 = A()
        assert len(events) == 4, len(events)
        assert a.derived_a is a.a1

        a.a2 = A()
        # Should not emit DerivedSet
        assert len(events) == 5, len(events)
        assert events[4].property is A.a2

        del events[:]
        old_a1 = a.a1
        del a.a1
        assert len(events) == 2, len(events)
        assert events[0].property is A.derived_a
        assert events[0].new_value is a.a2, "{} {} {}".format(
            a.a1, a.a2, events[3].new_value
        )
        assert events[0].old_value is old_a1, "{} {} {}".format(
            a.a1, a.a2, events[3].old_value
        )
        assert events[1].property is A.a1

        del events[:]
        old_a2 = a.a2
        del a.a2
        assert len(events) == 2, len(events)
        assert events[0].property is A.derived_a
        assert events[0].new_value is None, "{} {} {}".format(
            a.a1, a.a2, events[5].new_value
        )
        assert events[0].old_value is old_a2, "{} {} {}".format(
            a.a1, a.a2, events[5].old_value
        )
        assert events[1].property is A.a2

        del events[:]
        assert len(events) == 0, len(events)

        a.b1 = A()
        assert len(events) == 2
        assert events[0].property is A.derived_b
        assert events[1].property is A.b1

        a.b2 = A()
        assert len(events) == 4
        assert events[2].property is A.derived_b
        assert events[3].property is A.b2

        a.b2 = A()
        assert len(events) == 6
        assert events[4].property is A.derived_b
        assert events[5].property is A.b2

        a.b3 = A()
        assert len(events) == 8, len(events)
        assert events[6].property is A.derived_b
        assert events[7].property is A.b3

        # Add b3's value to b2, should not emit derived union event
        a.b2 = a.b3
        assert len(events) == 9, len(events)
        assert events[8].property is A.b2

        # Remove b3's value to b2
        del a.b2[a.b3]
        assert len(events) == 10, len(events)
        assert events[9].property is A.b2

        a.b3 = A()
        assert len(events) == 13, len(events)
        assert events[10].property is A.derived_b
        assert isinstance(events[10], DerivedDeleted), type(events[10])
        assert events[11].property is A.derived_b
        assert isinstance(events[11], DerivedAdded), type(events[11])
        assert events[12].property is A.b3

        del a.b3
        assert len(events) == 15, len(events)
        assert events[13].property is A.derived_b
        assert isinstance(events[13], DerivedDeleted), type(events[10])
        assert events[14].property is A.b3
    finally:
        Application.unregister_handler(handler)


@pytest.mark.skip
def test_redefine():
    from gaphor.application import Application

    class A(Element):
        is_unlinked = False

        def unlink(self):
            self.is_unlinked = True
            super().unlink()

    A.a = association("a", A, upper=1)

    A.a = redefine(A, "a", A, A.a)
    events = []

    @event_handler(AssociationUpdated)
    def handler(event, events=events):
        events.append(event)

    Application.register_handler(handler)
    try:
        a = A()
        a.a = A()
        assert len(events) == 2
        assert events[0].property is A.a, events[0].property
        assert events[1].property is A.a.original, events[1].property
    finally:
        Application.unregister_handler(handler)


@pytest.mark.skip
def test_redefine_subclass():
    class A(Element):
        is_unlinked = False

        def unlink():
            self.is_unlinked = True
            Element.unlink()

    A.a = association("a", A, upper=1)

    class B(A):
        pass

    B.b = redefine(B, "b", A, A.a)

    events = []

    @event_handler(AssociationUpdated)
    def handler(event, events=events):
        events.append(event)

    Application.register_handler(handler)
    try:
        a = A()
        a.a = A()
        # Only a.a changes, no B class involved
        assert len(events) == 1
        assert events[0].property is A.a, events[0].property
        # assert events[1].property is A.a.original, events[1].property
        del events[:]

        a = B()
        a.a = A()
        # Now events are sent for both association and redefine
        assert len(events) == 2
        assert events[0].property is B.b, events[0].property
        assert events[1].property is B.b.original, events[1].property
    finally:
        Application.unregister_handler(handler)



"""
Connection between two state machine vertices (state, pseudostate) using
transition.

To register connectors implemented in this module, it is imported in
gaphor.adapter package.
"""

from gaphor import UML
from gaphor.diagram.connectors import IConnect, RelationshipConnect
from gaphor.diagram.states.transition import TransitionItem
from gaphor.diagram.states.state import VertexItem
from gaphor.diagram.states.pseudostates import (
    InitialPseudostateItem,
    HistoryPseudostateItem,
)


class VertexConnect(RelationshipConnect):
    """
    Abstract relationship between two state vertices.
    """

    def reconnect(self, handle, port):
        self.reconnect_relationship(
            handle, UML.Transition.source, UML.Transition.target
        )

    def connect_subject(self, handle):
        relation = self.relationship_or_new(
            UML.Transition, UML.Transition.source, UML.Transition.target
        )
        self.line.subject = relation
        if relation.guard is None:
            relation.guard = self.line.model.create(UML.Constraint)


@IConnect.register(VertexItem, TransitionItem)
class TransitionConnect(VertexConnect):
    """Connect two state vertices using transition item."""

    def allow(self, handle, port):
        """
        Glue transition handle and vertex item. Guard from connecting
        transition's head with final state.
        """
        line = self.line
        subject = self.element.subject

        is_final = isinstance(subject, UML.FinalState)
        if (
            isinstance(subject, UML.State)
            and not is_final
            or handle is line.tail
            and is_final
        ):
            return super().allow(handle, port)
        else:
            return None


@IConnect.register(InitialPseudostateItem, TransitionItem)
class InitialPseudostateTransitionConnect(VertexConnect):
    """Connect initial pseudostate using transition item.

    It modifies InitialPseudostateItem._connected attribute to disallow
    connection of more than one transition.
    """

    def allow(self, handle, port):
        """
        Glue to initial pseudostate with transition's head and when there are
        no transitions connected.
        """
        line = self.line
        element = self.element
        subject = element.subject

        # Check if no other items are connected
        connections = self.canvas.get_connections(connected=element)
        connected_items = [
            c
            for c in connections
            if isinstance(c.item, TransitionItem) and c.item is not line
        ]
        if handle is line.head and not any(connected_items):
            return super().allow(handle, port)
        else:
            return None


@IConnect.register(HistoryPseudostateItem, TransitionItem)
class HistoryPseudostateTransitionConnect(VertexConnect):
    """Connect history pseudostate using transition item.

    It modifies InitialPseudostateItem._connected attribute to disallow
    connection of more than one transition.
    """

    def allow(self, handle, port):
        """
        """
        return super().allow(handle, port)



""" Event management system.

This module provides API for event management. There are two APIs provided:

* Global event management API: subscribe, unsubscribe, handle.
* Local event management API: Manager

If you run only one instance of your application per Python
interpreter you can stick with global API, but if you want to have
more than one application instances running inside one interpreter and
to have different configurations for them -- you should use local API
and have one instance of Manager object per application instance.
"""

from typing import Callable, Set, Type

from gaphor.misc.generic.registry import Registry, TypeAxis


__all__ = "Manager"

Event = object
Handler = Callable[[object], None]
HandlerSet = Set[Handler]


class Manager:
    """ Event manager

    Provides API for subscribing for and firing events. There's also global
    event manager instantiated at module level with functions
    :func:`.subscribe`, :func:`.handle` and decorator :func:`.subscriber` aliased
    to corresponding methods of class.
    """

    def __init__(self) -> None:
        axes = (("event_type", TypeAxis()),)
        self.registry = Registry[HandlerSet](*axes)

    def subscribe(self, handler: Handler, event_type: Type[Event]) -> None:
        """ Subscribe ``handler`` to specified ``event_type``"""
        handler_set = self.registry.get_registration(event_type)
        if handler_set is None:
            handler_set = self._register_handler_set(event_type)
        handler_set.add(handler)

    def unsubscribe(self, handler: Handler, event_type: Type[Event]) -> None:
        """ Unsubscribe ``handler`` from ``event_type``"""
        handler_set = self.registry.get_registration(event_type)
        if handler_set and handler in handler_set:
            handler_set.remove(handler)

    def handle(self, event: Event) -> None:
        """ Fire ``event``

        All subscribers will be executed with no determined order.
        """
        handler_sets = self.registry.query(event)
        for handler_set in handler_sets:
            if handler_set:
                for handler in set(handler_set):
                    handler(event)

    def _register_handler_set(self, event_type: Type[Event]) -> HandlerSet:
        """ Register new handler set for ``event_type``.
        """
        handler_set: HandlerSet = set()
        self.registry.register(handler_set, event_type)
        return handler_set

    def subscriber(self, event_type: Type[Event]) -> Callable[[Handler], Handler]:
        """ Decorator for subscribing handlers

        Works like this:

            >>> mymanager = Manager()
            >>> class MyEvent():
            ...     pass
            >>> @mymanager.subscriber(MyEvent)
            ... def mysubscriber(evt):
            ...     # handle event
            ...     return

            >>> mymanager.handle(MyEvent())

        """

        def registrator(func: Handler) -> Handler:
            self.subscribe(func, event_type)
            return func

        return registrator



import unittest

from gaphas.examples import Box

from gaphor import UML
from gaphor.application import Application
from gaphor.diagram.general.comment import CommentItem
from gaphor.ui.mainwindow import DiagramPage


class DiagramPageTestCase(unittest.TestCase):
    def setUp(self):
        Application.init(
            services=[
                "event_manager",
                "component_registry",
                "element_factory",
                "main_window",
                "properties",
                "namespace",
                "diagrams",
                "toolbox",
                "elementeditor",
                "export_menu",
                "tools_menu",
            ]
        )
        main_window = Application.get_service("main_window")
        main_window.open()
        self.element_factory = Application.get_service("element_factory")
        self.diagram = self.element_factory.create(UML.Diagram)
        self.page = DiagramPage(
            self.diagram,
            Application.get_service("event_manager"),
            self.element_factory,
            Application.get_service("properties"),
        )
        self.page.construct()
        assert self.page.diagram == self.diagram
        assert self.page.view.canvas == self.diagram.canvas
        assert len(self.element_factory.lselect()) == 1

    def tearDown(self):
        self.page.close()
        del self.page
        self.diagram.unlink()
        del self.diagram
        Application.shutdown()
        assert len(self.element_factory.lselect()) == 0

    def test_creation(self):
        pass

    def test_placement(self):
        box = Box()
        self.diagram.canvas.add(box)
        self.diagram.canvas.update_now()
        self.page.view.request_update([box])

        comment = self.diagram.create(
            CommentItem, subject=self.element_factory.create(UML.Comment)
        )
        assert len(self.element_factory.lselect()) == 2



from gi.repository import Gtk
from gaphor.ui.preferences import Preferences


class MainWindowMock:
    def __init__(self):
        self.window = Gtk.Window.new(Gtk.WindowType.TOPLEVEL)


def test_preferences_window():
    prefs = Preferences(MainWindowMock(), {})

    dialog = prefs.open()
    action_group = dialog.get_action_group("pref")

    assert (
        action_group.lookup_action("hand-drawn-style").get_state().get_boolean()
        == False
    )


def test_with_hand_drawn_style_enabled():
    prefs = Preferences(MainWindowMock(), {"diagram.sloppiness": 0.5})

    dialog = prefs.open()
    action_group = dialog.get_action_group("pref")

    assert (
        action_group.lookup_action("hand-drawn-style").get_state().get_boolean() == True
    )



"""
Comment and comment line items connection adapters tests.
"""

from gaphor import UML
from gaphor.tests import TestCase
from gaphor.diagram.general.comment import CommentItem
from gaphor.diagram.general.commentline import CommentLineItem
from gaphor.diagram.usecases.actor import ActorItem
from gaphor.diagram.classes.klass import ClassItem
from gaphor.diagram.classes.generalization import GeneralizationItem


class CommentLineTestCase(TestCase):

    services = TestCase.services + ["sanitizer"]

    # NOTE: Still have to test what happens if one Item at the CommentLineItem
    #       end is removed, while the item still has references and is not
    #       removed itself.

    def test_commentline_annotated_element(self):
        """Test comment line item annotated element creation
        """
        comment = self.create(CommentItem, UML.Comment)
        line = self.create(CommentLineItem)

        self.connect(line, line.head, comment)
        # connected, but no annotated element yet
        self.assertTrue(self.get_connected(line.head) is not None)
        assert not comment.subject.annotatedElement

    def test_commentline_same_comment_glue(self):
        """Test comment line item gluing to already connected comment item."""

        comment = self.create(CommentItem, UML.Comment)
        line = self.create(CommentLineItem)

        self.connect(line, line.head, comment)
        glued = self.allow(line, line.tail, comment)
        assert not glued

    def test_commentline_element_connect(self):
        """Test comment line connecting to comment and actor items.
        """
        comment = self.create(CommentItem, UML.Comment)
        line = self.create(CommentLineItem)
        ac = self.create(ActorItem, UML.Actor)

        self.connect(line, line.head, comment)
        self.connect(line, line.tail, ac)
        assert self.get_connected(line.tail) is ac
        assert 1 == len(comment.subject.annotatedElement)
        assert ac.subject in comment.subject.annotatedElement

    def test_commentline_element_reconnect(self):
        """Test comment line connecting to comment and actor items.
        """
        comment = self.create(CommentItem, UML.Comment)
        line = self.create(CommentLineItem)
        ac = self.create(ActorItem, UML.Actor)

        self.connect(line, line.head, comment)
        self.connect(line, line.tail, ac)
        assert self.get_connected(line.tail) is ac
        assert 1 == len(comment.subject.annotatedElement)
        assert ac.subject in comment.subject.annotatedElement

        ac2 = self.create(ActorItem, UML.Actor)
        # ac.canvas.disconnect_item(line, line.tail)
        self.disconnect(line, line.tail)
        self.connect(line, line.tail, ac2)
        assert self.get_connected(line.tail) is ac2
        assert 1 == len(comment.subject.annotatedElement)
        assert ac2.subject in comment.subject.annotatedElement

    def test_commentline_element_disconnect(self):
        """Test comment line connecting to comment and disconnecting actor item.
        """
        comment = self.create(CommentItem, UML.Comment)
        line = self.create(CommentLineItem)
        ac = self.create(ActorItem, UML.Actor)

        self.connect(line, line.head, comment)
        self.connect(line, line.tail, ac)

        assert self.get_connected(line.tail) is ac

        self.disconnect(line, line.tail)
        assert not self.get_connected(line.tail) is ac

    def test_commentline_unlink(self):
        """Test comment line unlinking.
        """
        clazz = self.create(ClassItem, UML.Class)
        comment = self.create(CommentItem, UML.Comment)
        line = self.create(CommentLineItem)

        self.connect(line, line.head, comment)
        self.connect(line, line.tail, clazz)
        assert clazz.subject in comment.subject.annotatedElement
        assert comment.subject in clazz.subject.ownedComment

        assert line.canvas

        # FixMe: This should invoke the disconnect handler of the line's
        #  handles.

        line.unlink()

        assert not line.canvas
        assert clazz.subject not in comment.subject.annotatedElement
        assert comment.subject not in clazz.subject.ownedComment
        assert (
            len(comment.subject.annotatedElement) == 0
        ), comment.subject.annotatedElement
        assert len(clazz.subject.ownedComment) == 0, clazz.subject.ownedComment

    def test_commentline_element_unlink(self):
        """Test comment line unlinking using a class item.
        """
        clazz = self.create(ClassItem, UML.Class)
        comment = self.create(CommentItem, UML.Comment)
        line = self.create(CommentLineItem)

        self.connect(line, line.head, comment)
        self.connect(line, line.tail, clazz)
        assert clazz.subject in comment.subject.annotatedElement
        assert comment.subject in clazz.subject.ownedComment

        assert line.canvas

        clazz_subject = clazz.subject

        # FixMe: This should invoke the disconnect handler of the line's
        #  handles.

        clazz.unlink()

        assert not clazz.canvas
        assert line.canvas
        assert not comment.subject.annotatedElement
        assert len(clazz_subject.ownedComment) == 0

    def test_commentline_relationship_unlink(self):
        """Test comment line to a relationship item connection and unlink.

        Demonstrates defect #103.
        """
        clazz1 = self.create(ClassItem, UML.Class)
        clazz2 = self.create(ClassItem, UML.Class)
        gen = self.create(GeneralizationItem)

        self.connect(gen, gen.head, clazz1)
        self.connect(gen, gen.tail, clazz2)

        assert gen.subject

        # now, connect comment to a generalization (relationship)
        comment = self.create(CommentItem, UML.Comment)
        line = self.create(CommentLineItem)
        self.connect(line, line.head, comment)
        self.connect(line, line.tail, gen)

        assert gen.subject in comment.subject.annotatedElement
        assert comment.subject in gen.subject.ownedComment

        gen.unlink()

        assert not comment.subject.annotatedElement
        assert gen.subject is None

    def test_commentline_linked_to_same_element_twice(self):
        """
        It is not allowed to create two commentlines between the same elements.
        """
        clazz = self.create(ClassItem, UML.Class)

        # now, connect comment to a generalization (relationship)
        comment = self.create(CommentItem, UML.Comment)
        line1 = self.create(CommentLineItem)
        self.connect(line1, line1.head, comment)
        self.connect(line1, line1.tail, clazz)

        assert clazz.subject in comment.subject.annotatedElement
        assert comment.subject in clazz.subject.ownedComment

        # Now add another line

        line2 = self.create(CommentLineItem)
        self.connect(line2, line2.head, comment)

        assert not self.allow(line2, line2.tail, clazz)


# vim: sw=4:et:ai



"""
Interaction diagram item.
"""

from gaphor import UML
from gaphor.UML.modelfactory import stereotypes_str
from gaphor.diagram.presentation import ElementPresentation, Named
from gaphor.diagram.shapes import Box, EditableText, Text, draw_highlight
from gaphor.diagram.support import represents
from gaphor.diagram.text import TextAlign, VerticalAlign


@represents(UML.Interaction)
class InteractionItem(ElementPresentation, Named):
    def __init__(self, id=None, model=None):
        super().__init__(id, model)

        self.shape = Box(
            Box(
                Text(
                    text=lambda: stereotypes_str(self.subject),
                    style={
                        "min-width": 0,
                        "min-height": 0,
                        "text-align": TextAlign.LEFT,
                    },
                ),
                EditableText(
                    text=lambda: self.subject.name or "",
                    style={"text-align": TextAlign.LEFT},
                ),
                style={"padding": (4, 4, 4, 4)},
            ),
            style={
                "min-width": 150,
                "min-height": 100,
                "vertical-align": VerticalAlign.TOP,
            },
            draw=draw_interaction,
        )

        self.watch("subject[NamedElement].name")
        self.watch("subject.appliedStereotype.classifier.name")


def draw_interaction(box, context, bounding_box):
    cr = context.cairo
    cr.rectangle(0, 0, bounding_box.width, bounding_box.height)
    draw_highlight(context)
    # draw pentagon
    w, h = box.sizes[0]
    h2 = h / 2.0
    cr.move_to(0, h)
    cr.line_to(w - 4, h)
    cr.line_to(w, h2)
    cr.line_to(w, 0)
    cr.stroke()



import pytest

from gaphor import UML
from gaphor.application import Application
from gaphor.ui.event import DiagramOpened
from gaphor.ui.abc import UIComponent


@pytest.fixture
def application():
    Application.init(
        services=[
            "event_manager",
            "component_registry",
            "element_factory",
            "properties",
            "main_window",
            "namespace",
            "diagrams",
            "toolbox",
            "elementeditor",
            "export_menu",
            "tools_menu",
        ]
    )
    yield Application
    Application.shutdown()


def get_current_diagram(app):
    return app.component_registry.get(UIComponent, "diagrams").get_current_diagram()


def test_creation(application):
    # MainWindow should be created as resource
    main_w = application.get_service("main_window")
    main_w.open()
    assert get_current_diagram(application) is None


def test_show_diagram(application):
    element_factory = application.get_service("element_factory")
    diagram = element_factory.create(UML.Diagram)
    main_w = application.get_service("main_window")
    main_w.open()
    event_manager = application.get_service("event_manager")
    event_manager.handle(DiagramOpened(diagram))
    assert get_current_diagram(application) == diagram



import pytest
from gaphor import UML
from gaphor.UML.elementfactory import ElementFactory
from gaphor.services.eventmanager import EventManager
from gaphor.diagram.general.generalpropertypages import CommentItemPropertyPage


@pytest.fixture
def element_factory():
    return ElementFactory(EventManager())


def test_property_page_construction(element_factory):
    comment = element_factory.create(UML.Comment)
    prop_page = CommentItemPropertyPage(comment)

    widget = prop_page.construct()
    text_view = widget.get_children()[-1]

    assert text_view



from gaphor.diagram.editors import Editor, AbstractEditor
from gaphor.diagram.general.comment import CommentItem


@Editor.register(CommentItem)
class CommentItemEditor(AbstractEditor):
    """Text edit support for Comment item."""

    def __init__(self, item):
        self._item = item

    def is_editable(self, x, y):
        return True

    def get_text(self):
        return self._item.subject.body

    def update_text(self, text):
        self._item.subject.body = text

    def key_pressed(self, pos, key):
        pass



"""
The file service is responsible for loading and saving the user data.
"""

from typing import Optional
import logging

from gi.repository import Gtk
import urllib.parse
from gaphor import UML
from gaphor.core import _, action, event_handler
from gaphor.abc import Service, ActionProvider
from gaphor.misc.errorhandler import error_handler
from gaphor.misc.gidlethread import GIdleThread, Queue
from gaphor.misc.xmlwriter import XMLWriter
from gaphor.storage import storage, verify
import gaphor.ui
from gaphor.ui.event import FileLoaded, FileSaved, WindowClosed
from gaphor.ui.filedialog import FileDialog
from gaphor.ui.questiondialog import QuestionDialog
from gaphor.ui.statuswindow import StatusWindow

DEFAULT_EXT = ".gaphor"
MAX_RECENT = 10

log = logging.getLogger(__name__)


class FileManager(Service, ActionProvider):
    """
    The file service, responsible for loading and saving Gaphor models.
    """

    def __init__(self, event_manager, element_factory, main_window, properties):
        """File manager constructor.  There is no current filename yet."""
        self.event_manager = event_manager
        self.element_factory = element_factory
        self.main_window = main_window
        self.properties = properties
        self._filename = None

        event_manager.subscribe(self._on_window_close)

    def shutdown(self):
        """Called when shutting down the file manager service."""
        self.event_manager.unsubscribe(self._on_window_close)

    def get_filename(self):
        """Return the current file name.  This method is used by the filename
        property."""
        return self._filename

    def set_filename(self, filename):
        """Sets the current file name.  This method is used by the filename
        property. Setting the current filename will update the recent file
        list."""

        if filename != self._filename:
            self._filename = filename

    filename = property(get_filename, set_filename)

    def load(self, filename):
        """Load the Gaphor model from the supplied file name.  A status window
        displays the loading progress.  The load generator updates the progress
        queue.  The loader is passed to a GIdleThread which executes the load
        generator.  If loading is successful, the filename is set."""

        queue = Queue()
        status_window: Optional[StatusWindow]
        try:
            main_window = self.main_window
            status_window = StatusWindow(
                _("Loading..."),
                _("Loading model from %s") % filename,
                parent=main_window.window,
                queue=queue,
            )
        except:
            log.warning("Could not create status window, proceding without.")
            status_window = None

        try:
            loader = storage.load_generator(
                filename.encode("utf-8"), self.element_factory
            )
            worker = GIdleThread(loader, queue)

            worker.start()
            worker.wait()

            if worker.error:
                worker.reraise()

            self.filename = filename
            self.event_manager.handle(FileLoaded(self, filename))
        except:
            error_handler(
                message=_("Error while loading model from file %s") % filename
            )
            raise
        finally:
            if status_window is not None:
                status_window.destroy()

    def verify_orphans(self):
        """Verify that no orphaned elements are saved.  This method checks
        of there are any orphan references in the element factory.  If orphans
        are found, a dialog is displayed asking the user if it is OK to
        unlink them."""

        orphans = verify.orphan_references(self.element_factory)

        if orphans:
            main_window = self.main_window

            dialog = QuestionDialog(
                _(
                    "The model contains some references"
                    " to items that are not maintained."
                    " Do you want to clean this before"
                    " saving the model?"
                ),
                parent=main_window.window,
            )

            answer = dialog.answer
            dialog.destroy()

            if not answer:
                for orphan in orphans:
                    orphan.unlink()

    def verify_filename(self, filename):
        """Verify that the supplied filename is using the proper default
        extension.  If not, the extension is added to the filename
        and returned."""

        if not filename.endswith(DEFAULT_EXT):
            filename = filename + DEFAULT_EXT

        return filename

    def save(self, filename):
        """Save the current UML model to the specified file name.  Before
        writing the model file, this will verify that there are no orphan
        references.  It will also verify that the filename has the correct
        extension.  A status window is displayed while the GIdleThread
        is executed.  This thread actually saves the model."""

        if not filename or not len(filename):
            return

        self.verify_orphans()
        filename = self.verify_filename(filename)

        main_window = self.main_window
        queue = Queue()
        status_window = StatusWindow(
            _("Saving..."),
            _("Saving model to %s") % filename,
            parent=main_window.window,
            queue=queue,
        )
        try:
            with open(filename.encode("utf-8"), "w") as out:
                saver = storage.save_generator(XMLWriter(out), self.element_factory)
                worker = GIdleThread(saver, queue)
                worker.start()
                worker.wait()

            if worker.error:
                worker.reraise()

            self.filename = filename
            self.event_manager.handle(FileSaved(self, filename))
        except:
            error_handler(message=_("Error while saving model to file %s") % filename)
            raise
        finally:
            status_window.destroy()

    def _open_dialog(self, title):
        """Open a file chooser dialog to select a model
        file to open."""

        filesel = Gtk.FileChooserDialog(
            title=title,
            action=Gtk.FileChooserAction.OPEN,
            buttons=(
                Gtk.STOCK_CANCEL,
                Gtk.ResponseType.CANCEL,
                Gtk.STOCK_OPEN,
                Gtk.ResponseType.OK,
            ),
        )
        filesel.set_transient_for(self.main_window.window)

        filter = Gtk.FileFilter()
        filter.set_name("Gaphor models")
        filter.add_pattern("*.gaphor")
        filesel.add_filter(filter)

        filter = Gtk.FileFilter()
        filter.set_name("All files")
        filter.add_pattern("*")
        filesel.add_filter(filter)

        if self.filename:
            filesel.set_current_name(self.filename)

        response = filesel.run()
        filename = filesel.get_filename()
        filesel.destroy()
        if not filename or response != Gtk.ResponseType.OK:
            return
        return filename

    @action(name="file-new", shortcut="<Primary>n")
    def action_new(self):
        """The new model menu action.  This action will create a new
        UML model.  This will trigger a FileManagerStateChange event."""

        element_factory = self.element_factory
        main_window = self.main_window

        if element_factory.size():
            dialog = QuestionDialog(
                _(
                    "Opening a new model will flush the"
                    " currently loaded model.\nAny changes"
                    " made will not be saved. Do you want to"
                    " continue?"
                ),
                parent=main_window.window,
            )

            answer = dialog.answer
            dialog.destroy()

            if not answer:
                return

        element_factory.flush()
        with element_factory.block_events():
            model = element_factory.create(UML.Package)
            model.name = _("New model")
            diagram = element_factory.create(UML.Diagram)
            diagram.package = model
            diagram.name = _("main")
        self.filename = None
        element_factory.model_ready()

        # main_window.select_element(diagram)
        # main_window.show_diagram(diagram)

    @action(name="file-new-template")
    def action_new_from_template(self):
        """This menu action opens the new model from template dialog."""

        filters = [
            {"name": _("Gaphor Models"), "pattern": "*.gaphor"},
            {"name": _("All Files"), "pattern": "*"},
        ]

        file_dialog = FileDialog(_("New Gaphor Model From Template"), filters=filters)

        filename = file_dialog.selection

        file_dialog.destroy()

        log.debug(filename)

        if filename:
            self.load(filename)
            self.filename = None

    @action(name="file-open", shortcut="<Primary>o")
    def action_open(self):
        """This menu action opens the standard model open dialog."""

        filters = [
            {"name": _("Gaphor Models"), "pattern": "*.gaphor"},
            {"name": _("All Files"), "pattern": "*"},
        ]

        file_dialog = FileDialog(_("Open Gaphor Model"), filters=filters)

        filename = file_dialog.selection

        file_dialog.destroy()

        log.debug(filename)

        if filename:
            self.load(filename)

    @action(name="file-open-recent")
    def action_open_recent(self, file_url: str):
        parsed_url = urllib.parse.urlparse(file_url)
        path = parsed_url.path
        self.load(path)

    @action(name="file-save", shortcut="<Primary>s")
    def action_save(self):
        """
        Save the file. Depending on if there is a file name, either perform
        the save directly or present the user with a save dialog box.

        Returns True if the saving actually succeeded.
        """

        filename = self.filename

        if filename:
            self.save(filename)
            return True
        else:
            return self.action_save_as()

    @action(name="file-save-as", shortcut="<Primary><Shift>s")
    def action_save_as(self):
        """
        Save the model in the element_factory by allowing the
        user to select a file name.

        Returns True if the saving actually happened.
        """

        file_dialog = FileDialog(
            _("Save Gaphor Model As"), action="save", filename=self.filename
        )

        filename = file_dialog.selection

        file_dialog.destroy()

        if filename:
            self.save(filename)
            return True

        return False

    @action(name="app.quit", shortcut="<Primary>q")
    def file_quit(self):
        """
        Ask user to close window if the model has changed.
        The user is asked to either discard the changes, keep the
        application running or save the model and quit afterwards.
        """
        if self.main_window.model_changed:
            dialog = Gtk.MessageDialog(
                self.main_window.window,
                Gtk.DialogFlags.MODAL | Gtk.DialogFlags.DESTROY_WITH_PARENT,
                Gtk.MessageType.WARNING,
                Gtk.ButtonsType.NONE,
                _("Save changed to your model before closing?"),
            )
            dialog.format_secondary_text(
                _("If you close without saving, your changes will be discarded.")
            )
            dialog.add_buttons(
                _("Close _without saving"),
                Gtk.ResponseType.REJECT,
                Gtk.STOCK_CANCEL,
                Gtk.ResponseType.CANCEL,
                Gtk.STOCK_SAVE,
                Gtk.ResponseType.YES,
            )
            dialog.set_default_response(Gtk.ResponseType.YES)
            response = dialog.run()
            dialog.destroy()

            if response == Gtk.ResponseType.YES:
                saved = self.action_save()
                if saved:
                    gaphor.ui.quit()
            if response == Gtk.ResponseType.REJECT:
                gaphor.ui.quit()
        else:
            gaphor.ui.quit()

    @event_handler(WindowClosed)
    def _on_window_close(self, event):
        self.file_quit()



import abc


class Service(metaclass=abc.ABCMeta):
    """
    Base interface for all services in Gaphor.
    """

    @abc.abstractmethod
    def shutdown(self) -> None:
        """
        Shutdown the services, free resources.
        """


class ActionProvider(metaclass=abc.ABCMeta):
    """
    An action provider is a special service that provides actions
    (see gaphor/action.py) and the accompanying XML for the UI manager.
    """



"""This module has a generic FileDialog class that is used to open
or save files."""

from gi.repository import Gtk


class FileDialog:
    """This is a file dialog that is used to open or save a file."""

    def __init__(
        self,
        title,
        filename=None,
        action="open",
        parent=None,
        multiple=False,
        filters=[],
    ):
        """Initialize the file dialog.  The title parameter is the title
        displayed in the dialog.  The filename parameter will set the current
        file name in the dialog.  The action is either open or save and changes
        the buttons displayed.  If the parent window parameter is supplied,
        the file dialog is set to be transient for that window.  The multiple
        parameter should be set to true if multiple files can be opened at once.
        This means that a list of filenames instead of a single filename string
        will be returned by the selection property."""

        self.multiple = multiple

        if action == "open":
            action = Gtk.FileChooserAction.OPEN
        else:
            action = Gtk.FileChooserAction.SAVE

        self.dialog = Gtk.FileChooserNative(title=title, action=action)

        if parent:
            self.dialog.set_transient_for(parent)

        if filename:
            self.dialog.set_current_name(filename)

    def get_selection(self):
        """Return the selected file or files from the dialog.  This is used
        by the selection property."""

        response = self.dialog.run()

        if response == Gtk.ResponseType.ACCEPT:
            if self.multiple:
                selection = self.dialog.get_filenames()
            else:
                selection = self.dialog.get_filename()
        else:
            selection = ""

        return selection

    def destroy(self):
        """Destroy the GTK dialog."""

        self.dialog.destroy()

    selection = property(get_selection)



"""
Stereotype property page.
"""

from gi.repository import Gtk

from gaphor import UML
from gaphor.core import _, transactional
from gaphor.diagram.propertypages import PropertyPages, PropertyPageBase


def create_stereotype_tree_view(model, toggle_stereotype, set_slot_value):
    """
    Create a tree view for an editable tree model.

    :Parameters:
     model
        Model, for which tree view is created.
    """
    tree_view = Gtk.TreeView.new_with_model(model)

    # Stereotype/Attributes
    col = Gtk.TreeViewColumn.new()
    col.set_title("{} / {}".format(_("Stereotype"), _("Attribute")))
    col.set_expand(True)
    renderer = Gtk.CellRendererToggle()
    renderer.set_property("active", True)
    renderer.set_property("activatable", True)
    renderer.connect("toggled", toggle_stereotype, model, 2)
    col.pack_start(renderer, False)
    col.add_attribute(renderer, "active", 2)

    def show_checkbox(column, cell, model, iter, data):
        # value = model.get_value(iter, 4)
        # cell.set_property('active', value is not None)
        value = model.get_value(iter, 3)
        cell.set_property("visible", isinstance(value, UML.Stereotype))

    col.set_cell_data_func(renderer, show_checkbox)

    renderer = Gtk.CellRendererText.new()
    renderer.set_property("editable", False)
    renderer.set_property("is-expanded", True)
    col.pack_start(renderer, False)
    col.add_attribute(renderer, "text", 0)
    tree_view.append_column(col)

    # TODO: use col.set_cell_data_func(renderer, func, None) to toggle visibility
    # Value
    renderer = Gtk.CellRendererText()
    renderer.set_property("is-expanded", True)
    renderer.connect("edited", set_slot_value, model, 1)
    col = Gtk.TreeViewColumn(_("Value"), renderer, text=1)
    col.set_expand(True)

    def set_editable(column, cell, model, iter, data):
        value = model.get_value(iter, 4)
        cell.set_property("editable", bool(value))

    col.set_cell_data_func(renderer, set_editable)
    tree_view.append_column(col)

    # tree_view.connect('key_press_event', remove_on_keypress)
    # tree_view.connect('key_press_event', swap_on_keypress)

    return tree_view


@PropertyPages.register(UML.Element)
class StereotypePage(PropertyPageBase):

    order = 40
    name = "Stereotypes"

    def __init__(self, item):
        self.item = item
        self.size_group = Gtk.SizeGroup.new(Gtk.SizeGroupMode.HORIZONTAL)

    def construct(self):
        page = Gtk.VBox()
        subject = self.item.subject
        if subject is None:
            return None

        stereotypes = UML.model.get_stereotypes(subject)
        if not stereotypes:
            return None

        # show stereotypes attributes toggle
        if hasattr(self.item, "show_stereotypes"):
            hbox = Gtk.HBox()
            label = Gtk.Label(label="")
            hbox.pack_start(label, False, True, 0)
            button = Gtk.CheckButton(label=_("Show stereotypes attributes"))
            button.set_active(self.item.show_stereotypes)
            button.connect("toggled", self._on_show_stereotypes_change)
            hbox.pack_start(button, True, True, 0)
            page.pack_start(hbox, False, True, 0)

        # stereotype attributes
        # self.model = StereotypeAttributes(self.item.subject)
        self.model = Gtk.TreeStore.new([str, str, bool, object, object, object])
        tree_view = create_stereotype_tree_view(
            self.model, self._toggle_stereotype, self._set_value
        )
        page.pack_start(tree_view, True, True, 0)

        page.show_all()

        self.refresh()
        return page

    @transactional
    def _on_show_stereotypes_change(self, button):
        self.item.show_stereotypes = button.get_active()

    def refresh(self):
        self.model.clear()
        subject = self.item.subject
        stereotypes = UML.model.get_stereotypes(subject)
        instances = subject.appliedStereotype

        # shortcut map stereotype -> slot (InstanceSpecification)
        slots = {}
        for obj in instances:
            for slot in obj.slot:
                slots[slot.definingFeature] = slot

        for st in stereotypes:
            for obj in instances:
                if st in obj.classifier:
                    break
            else:
                obj = None

            parent = self.model.append(None, (st.name, "", bool(obj), st, None, None))

            if obj:
                for attr in st.ownedAttribute:
                    if not attr.association:
                        slot = slots.get(attr)
                        value = slot.value if slot else ""
                        data = (attr.name, value, True, attr, obj, slot)
                        # print 'data', data
                        self.model.append(parent, data)
            else:
                for attr in st.ownedAttribute:
                    if not attr.association:
                        data = (attr.name, "", False, attr, None, None)
                        # print 'no data', data
                        self.model.append(parent, data)

    @transactional
    def _set_value(self, renderer, path, value, model, col=0):
        iter = model.get_iter(path)
        self.set_slot_value(iter, value)

    @transactional
    def _toggle_stereotype(self, renderer, path, model, col):
        iter = model.get_iter(path)
        self.select_stereotype(iter)

    def select_stereotype(self, iter):
        """
        Select the stereotype.
        """
        path = self.model.get_path(iter)
        row = self.model[path]
        name, old_value, is_applied, stereotype, _, _ = row
        value = not is_applied

        subject = self.item.subject
        if value:
            UML.model.apply_stereotype(subject, stereotype)
        else:
            UML.model.remove_stereotype(subject, stereotype)

        row[2] = value

        # TODO: change refresh in a refresh of the data model, rather than a clear-refresh
        self.refresh()

    def set_slot_value(self, iter, value):
        """
        Set value of stereotype property applied to an UML element.

        Slot is created if instance Create valuChange value of instance spe
        """
        path = self.model.get_path(iter)
        row = self.model[path]
        name, old_value, is_applied, attr, obj, slot = row
        if isinstance(attr, UML.Stereotype):
            return  # don't edit stereotype rows

        if slot is None and not value:
            return  # nothing to do and don't create slot without value

        if slot is None:
            slot = UML.model.add_slot(obj, attr)

        assert slot

        if value:
            slot.value = value
        else:
            # no value, then remove slot
            del obj.slot[slot]
            slot = None
            value = ""

        row[1] = value
        row[5] = slot


# vim:sw=4:et:ai



import ast

import gaphas
from gaphas.geometry import Rectangle, distance_rectangle_point

from gaphor.UML.presentation import Presentation, S
from gaphor.diagram.text import text_point_at_line, TextAlign


class Named:
    """
    Marker for any NamedElement presentations.
    """


class Classified(Named):
    """
    Marker for Classifier presentations.
    """


def from_package_str(item):
    """
    Display name space info when it is different, then diagram's or
    parent's namespace.
    """
    subject = item.subject
    canvas = item.canvas

    if not subject or not canvas:
        return False

    namespace = subject.namespace
    parent = canvas.get_parent(item)

    # if there is a parent (i.e. interaction)
    if parent and parent.subject and parent.subject.namespace is not namespace:
        return False

    return (
        f"(from {namespace.name})" if namespace is not canvas.diagram.namespace else ""
    )


# Note: the official documentation is using the terms "Shape" and "Edge" for element and line.


class ElementPresentation(Presentation[S], gaphas.Element):
    """
    Presentation for Gaphas Element (box-like) items.

    To create a shape (boxes, text), assign a shape to `self.shape`.
    If the shape can change, for example because styling needs to change, implement
    the method `update_shapes()` and set self.shape there.
    """

    width: int
    height: int

    def __init__(self, id=None, model=None, shape=None):
        super().__init__(id, model)
        self._shape = shape

    def _set_shape(self, shape):
        self._shape = shape
        self.request_update()

    shape = property(lambda s: s._shape, _set_shape)

    def update_shapes(self, event=None):
        """
        Updating the shape configuration, e.g. when extra elements have to
        be drawn or when styling changes.
        """

    def pre_update(self, context):
        cr = context.cairo
        self.min_width, self.min_height = self.shape.size(cr)

    def draw(self, context):
        self._shape.draw(context, Rectangle(0, 0, self.width, self.height))

    def setup_canvas(self):
        super().setup_canvas()
        self.subscribe_all()
        # Invoke here, since we do not receive events, unless we're attached to a canvas
        self.update_shapes()

    def teardown_canvas(self):
        self.unsubscribe_all()
        super().teardown_canvas()

    def save(self, save_func):
        save_func("matrix", tuple(self.matrix))
        for prop in ("width", "height"):
            save_func(prop, getattr(self, prop))
        super().save(save_func)

    def load(self, name, value):
        if name == "matrix":
            self.matrix = ast.literal_eval(value)
        elif name in ("width", "height"):
            setattr(self, name, ast.literal_eval(value))
        else:
            super().load(name, value)

    def postload(self):
        super().postload()
        self.update_shapes()


class LinePresentation(Presentation[S], gaphas.Line):
    def __init__(
        self,
        id=None,
        model=None,
        style={},
        shape_head=None,
        shape_middle=None,
        shape_tail=None,
    ):
        super().__init__(id, model)

        self._style = {"dash-style": (), "line-width": 2, **style}

        self.shape_head = shape_head
        self.shape_middle = shape_middle
        self.shape_tail = shape_tail

        self.fuzziness = 2
        self._shape_head_rect = None
        self._shape_middle_rect = None
        self._shape_tail_rect = None

    head = property(lambda self: self._handles[0])
    tail = property(lambda self: self._handles[-1])

    def _set_style(self, style):
        self._style.update(style)

    style = property(
        lambda self: self._style.__getitem__,
        _set_style,
        doc="""A line, contrary to an element, has some styling of it's own.""",
    )

    def post_update(self, context):
        def shape_bounds(shape, align):
            if shape:
                size = shape.size(cr)
                x, y = text_point_at_line(points, size, align)
                return Rectangle(x, y, *size)

        super().post_update(context)
        cr = context.cairo
        points = [h.pos for h in self.handles()]
        self._shape_head_rect = shape_bounds(self.shape_head, TextAlign.LEFT)
        self._shape_middle_rect = shape_bounds(self.shape_middle, TextAlign.CENTER)
        self._shape_tail_rect = shape_bounds(self.shape_tail, TextAlign.RIGHT)

    def point(self, pos):
        """Given a point (x, y) return the distance to the canvas item.
        """
        d0 = super().point(pos)
        ds = [
            distance_rectangle_point(shape, pos)
            for shape in (
                self._shape_head_rect,
                self._shape_middle_rect,
                self._shape_tail_rect,
            )
            if shape
        ]
        return min(d0, *ds) if ds else d0

    def draw(self, context):
        cr = context.cairo
        cr.set_line_width(self.style("line-width"))
        if self.style("dash-style"):
            cr.set_dash(self.style("dash-style"), 0)

        super().draw(context)

        for shape, rect in (
            (self.shape_head, self._shape_head_rect),
            (self.shape_middle, self._shape_middle_rect),
            (self.shape_tail, self._shape_tail_rect),
        ):
            if shape:
                shape.draw(context, rect)

    def setup_canvas(self):
        super().setup_canvas()
        self.subscribe_all()

    def teardown_canvas(self):
        self.unsubscribe_all()
        super().teardown_canvas()

    def save(self, save_func):
        def save_connection(name, handle):
            c = self.canvas.get_connection(handle)
            if c:
                save_func(name, c.connected, reference=True)

        super().save(save_func)
        save_func("matrix", tuple(self.matrix))
        for prop in ("orthogonal", "horizontal"):
            save_func(prop, getattr(self, prop))
        points = [tuple(map(float, h.pos)) for h in self.handles()]
        save_func("points", points)

        save_connection("head-connection", self.head)
        save_connection("tail-connection", self.tail)

    def load(self, name, value):
        if name == "matrix":
            self.matrix = ast.literal_eval(value)
        elif name == "points":
            points = ast.literal_eval(value)
            for x in range(len(points) - 2):
                h = self._create_handle((0, 0))
                self._handles.insert(1, h)
            for i, p in enumerate(points):
                self.handles()[i].pos = p

            # Update connection ports of the line. Only handles are saved
            # in Gaphor file therefore ports need to be recreated after
            # handles information is loaded.
            self._update_ports()

        elif name == "orthogonal":
            self._load_orthogonal = ast.literal_eval(value)
        elif name == "horizontal":
            self.horizontal = ast.literal_eval(value)
        elif name in ("head_connection", "head-connection"):
            self._load_head_connection = value
        elif name in ("tail_connection", "tail-connection"):
            self._load_tail_connection = value
        else:
            super().load(name, value)

    def postload(self):
        def get_sink(handle, item):

            hpos = self.canvas.get_matrix_i2i(self, item).transform_point(*handle.pos)
            port = None
            dist = 10e6
            for p in item.ports():
                pos, d = p.glue(hpos)
                if not port or d < dist:
                    port = p
                    dist = d

            return gaphas.aspect.ConnectionSink(item, port)

        def postload_connect(handle, item):
            connector = gaphas.aspect.Connector(self, handle)
            sink = get_sink(handle, item)
            connector.connect(sink)

        if hasattr(self, "_load_orthogonal"):
            # Ensure there are enough handles
            if self._load_orthogonal and len(self._handles) < 3:
                p0 = self._handles[-1].pos
                self._handles.insert(1, self._create_handle(p0))
            self.orthogonal = self._load_orthogonal
            del self._load_orthogonal

        # First update matrix and solve constraints (NE and SW handle are
        # lazy and are resolved by the constraint solver rather than set
        # directly.
        self.canvas.update_matrix(self)
        self.canvas.solver.solve()

        if hasattr(self, "_load_head_connection"):
            postload_connect(self.head, self._load_head_connection)
            del self._load_head_connection

        if hasattr(self, "_load_tail_connection"):
            postload_connect(self.tail, self._load_tail_connection)
            del self._load_tail_connection

        super().postload()



"""
Classes related adapter connection tests.
"""

from gaphor.tests import TestCase
from gaphor import UML
from gaphor.diagram.usecases.actor import ActorItem
from gaphor.diagram.classes.dependency import DependencyItem
from gaphor.diagram.classes.klass import ClassItem
from gaphor.diagram.classes.interface import InterfaceItem
from gaphor.diagram.classes.association import AssociationItem
from gaphor.diagram.classes.generalization import GeneralizationItem


class DependencyTestCase(TestCase):
    """
    Dependency item connection adapter tests.
    """

    def test_dependency_glue(self):
        """Test dependency glue to two actor items
        """
        actor1 = self.create(ActorItem, UML.Actor)
        actor2 = self.create(ActorItem, UML.Actor)
        dep = self.create(DependencyItem)

        glued = self.allow(dep, dep.head, actor1)
        assert glued

        self.connect(dep, dep.head, actor1)

        glued = self.allow(dep, dep.tail, actor2)
        assert glued

    def test_dependency_connect(self):
        """Test dependency connecting to two actor items
        """
        actor1 = self.create(ActorItem, UML.Actor)
        actor2 = self.create(ActorItem, UML.Actor)
        dep = self.create(DependencyItem)

        self.connect(dep, dep.head, actor1)
        self.connect(dep, dep.tail, actor2)

        assert dep.subject is not None
        assert isinstance(dep.subject, UML.Dependency)
        assert dep.subject in self.element_factory.select()

        hct = self.get_connected(dep.head)
        tct = self.get_connected(dep.tail)
        assert hct is actor1
        assert tct is actor2

        assert actor1.subject in dep.subject.supplier
        assert actor2.subject in dep.subject.client

    def test_dependency_reconnection(self):
        """Test dependency reconnection
        """
        a1 = self.create(ActorItem, UML.Actor)
        a2 = self.create(ActorItem, UML.Actor)
        a3 = self.create(ActorItem, UML.Actor)
        dep = self.create(DependencyItem)

        # connect: a1 -> a2
        self.connect(dep, dep.head, a1)
        self.connect(dep, dep.tail, a2)

        d = dep.subject

        # reconnect: a1 -> a3
        self.connect(dep, dep.tail, a3)

        assert d is dep.subject
        assert 1 == len(dep.subject.supplier)
        assert 1 == len(dep.subject.client)
        assert a1.subject in dep.subject.supplier
        assert a3.subject in dep.subject.client
        assert a2.subject not in dep.subject.client, dep.subject.client

    def test_dependency_disconnect(self):
        """Test dependency disconnecting using two actor items
        """
        actor1 = self.create(ActorItem, UML.Actor)
        actor2 = self.create(ActorItem, UML.Actor)
        dep = self.create(DependencyItem)

        self.connect(dep, dep.head, actor1)
        self.connect(dep, dep.tail, actor2)

        dep_subj = dep.subject
        self.disconnect(dep, dep.tail)

        assert dep.subject is None
        assert self.get_connected(dep.tail) is None
        assert dep_subj not in self.element_factory.select()
        assert dep_subj not in actor1.subject.supplierDependency
        assert dep_subj not in actor2.subject.clientDependency

    def test_dependency_reconnect(self):
        """Test dependency reconnection using two actor items
        """
        actor1 = self.create(ActorItem, UML.Actor)
        actor2 = self.create(ActorItem, UML.Actor)
        dep = self.create(DependencyItem)

        self.connect(dep, dep.head, actor1)
        self.connect(dep, dep.tail, actor2)

        dep_subj = dep.subject
        self.disconnect(dep, dep.tail)

        # reconnect
        self.connect(dep, dep.tail, actor2)

        assert dep.subject is not None
        assert dep.subject is not dep_subj  # the old subject has been deleted
        assert dep.subject in actor1.subject.supplierDependency
        assert dep.subject in actor2.subject.clientDependency
        # TODO: test with interface (usage) and component (realization)
        # TODO: test with multiple diagrams (should reuse existing relationships first)

    def test_multi_dependency(self):
        """Test multiple dependencies

        Dependency should appear in a new diagram, bound on a new
        dependency item.
        """
        actoritem1 = self.create(ActorItem, UML.Actor)
        actoritem2 = self.create(ActorItem, UML.Actor)
        actor1 = actoritem1.subject
        actor2 = actoritem2.subject
        dep = self.create(DependencyItem)

        self.connect(dep, dep.head, actoritem1)
        self.connect(dep, dep.tail, actoritem2)

        assert dep.subject
        assert 1 == len(actor1.supplierDependency)
        assert actor1.supplierDependency[0] is dep.subject
        assert 1 == len(actor2.clientDependency)
        assert actor2.clientDependency[0] is dep.subject

        # Do the same thing, but now on a new diagram:

        diagram2 = self.element_factory.create(UML.Diagram)
        actoritem3 = diagram2.create(ActorItem, subject=actor1)
        actoritem4 = diagram2.create(ActorItem, subject=actor2)
        dep2 = diagram2.create(DependencyItem)

        self.connect(dep2, dep2.head, actoritem3)
        cinfo = diagram2.canvas.get_connection(dep2.head)
        assert cinfo is not None
        assert cinfo.connected is actoritem3
        self.connect(dep2, dep2.tail, actoritem4)
        assert dep2.subject is not None
        assert 1 == len(actor1.supplierDependency)
        assert actor1.supplierDependency[0] is dep.subject
        assert 1 == len(actor2.clientDependency)
        assert actor2.clientDependency[0] is dep.subject

        assert dep.subject is dep2.subject

    def test_dependency_type_auto(self):
        """Test dependency type automatic determination
        """
        cls = self.create(ClassItem, UML.Class)
        iface = self.create(InterfaceItem, UML.Interface)
        dep = self.create(DependencyItem)

        assert dep.auto_dependency

        self.connect(dep, dep.tail, cls)  # connect client
        self.connect(dep, dep.head, iface)  # connect supplier

        assert dep.subject is not None
        assert isinstance(dep.subject, UML.Usage), dep.subject
        assert dep.subject in self.element_factory.select()


class GeneralizationTestCase(TestCase):
    """
    Generalization item connection adapter tests.
    """

    def test_glue(self):
        """Test generalization item gluing using two classes."""

        gen = self.create(GeneralizationItem)
        c1 = self.create(ClassItem, UML.Class)
        c2 = self.create(ClassItem, UML.Class)

        glued = self.allow(gen, gen.tail, c1)
        assert glued

        self.connect(gen, gen.tail, c1)
        assert self.get_connected(gen.tail) is c1
        assert gen.subject is None

        glued = self.allow(gen, gen.head, c2)
        assert glued

    def test_connection(self):
        """Test generalization item connection using two classes
        """
        gen = self.create(GeneralizationItem)
        c1 = self.create(ClassItem, UML.Class)
        c2 = self.create(ClassItem, UML.Class)

        self.connect(gen, gen.tail, c1)
        assert self.get_connected(gen.tail) is c1

        self.connect(gen, gen.head, c2)
        assert gen.subject is not None
        assert gen.subject.general is c2.subject
        assert gen.subject.specific is c1.subject

    def test_reconnection(self):
        """Test generalization item connection using two classes

        On reconnection a new Generalization is created.
        """
        gen = self.create(GeneralizationItem)
        c1 = self.create(ClassItem, UML.Class)
        c2 = self.create(ClassItem, UML.Class)

        self.connect(gen, gen.tail, c1)
        assert self.get_connected(gen.tail) is c1

        self.connect(gen, gen.head, c2)
        assert gen.subject is not None
        assert gen.subject.general is c2.subject
        assert gen.subject.specific is c1.subject

        # Now do the same on a new diagram:
        diagram2 = self.element_factory.create(UML.Diagram)
        c3 = diagram2.create(ClassItem, subject=c1.subject)
        c4 = diagram2.create(ClassItem, subject=c2.subject)
        gen2 = diagram2.create(GeneralizationItem)

        self.connect(gen2, gen2.head, c3)
        cinfo = diagram2.canvas.get_connection(gen2.head)
        assert cinfo is not None
        assert cinfo.connected is c3

        self.connect(gen2, gen2.tail, c4)
        assert gen.subject is not gen2.subject
        assert 1 == len(c1.subject.generalization)
        assert c1.subject.generalization[0] is gen.subject
        # self.assertEqual(1, len(actor2.clientDependency))
        # self.assertTrue(actor2.clientDependency[0] is dep.subject)

    def test_reconnection2(self):
        """Test reconnection of generalization
        """
        c1 = self.create(ClassItem, UML.Class)
        c2 = self.create(ClassItem, UML.Class)
        c3 = self.create(ClassItem, UML.Class)
        gen = self.create(GeneralizationItem)

        # connect: c1 -> c2
        self.connect(gen, gen.head, c1)
        self.connect(gen, gen.tail, c2)

        s = gen.subject

        # reconnect: c2 -> c3
        self.connect(gen, gen.tail, c3)

        assert s is gen.subject
        assert c1.subject is gen.subject.general
        assert c3.subject is gen.subject.specific
        assert c2.subject is not gen.subject.specific


class AssociationConnectorTestCase(TestCase):
    """
    Association item connection adapters tests.
    """

    def test_glue(self):
        """Test association item glue
        """
        asc = self.create(AssociationItem)
        c1 = self.create(ClassItem, UML.Class)
        c2 = self.create(ClassItem, UML.Class)

        glued = self.allow(asc, asc.head, c1)
        assert glued

        self.connect(asc, asc.head, c1)

        glued = self.allow(asc, asc.tail, c2)
        assert glued

    def test_connect(self):
        """Test association item connection
        """
        asc = self.create(AssociationItem)
        c1 = self.create(ClassItem, UML.Class)
        c2 = self.create(ClassItem, UML.Class)

        self.connect(asc, asc.head, c1)
        assert asc.subject is None  # no UML metaclass yet

        self.connect(asc, asc.tail, c2)
        assert asc.subject is not None

        # Diagram, Class *2, Property *2, Association
        self.assertEqual(6, len(list(self.element_factory.select())))
        assert asc.head_end.subject is not None
        assert asc.tail_end.subject is not None

    def test_reconnect(self):
        """Test association item reconnection
        """
        asc = self.create(AssociationItem)
        c1 = self.create(ClassItem, UML.Class)
        c2 = self.create(ClassItem, UML.Class)
        c3 = self.create(ClassItem, UML.Class)

        self.connect(asc, asc.head, c1)
        self.connect(asc, asc.tail, c2)
        UML.model.set_navigability(asc.subject, asc.tail_end.subject, True)

        a = asc.subject

        self.connect(asc, asc.tail, c3)

        assert a is asc.subject
        ends = [p.type for p in asc.subject.memberEnd]
        assert c1.subject in ends
        assert c3.subject in ends
        assert c2.subject not in ends
        assert asc.tail_end.subject.navigability

    def test_disconnect(self):
        """Test association item disconnection
        """
        asc = self.create(AssociationItem)
        c1 = self.create(ClassItem, UML.Class)
        c2 = self.create(ClassItem, UML.Class)

        self.connect(asc, asc.head, c1)
        assert asc.subject is None  # no UML metaclass yet

        self.connect(asc, asc.tail, c2)
        assert asc.subject is not None

        self.disconnect(asc, asc.head)

        # after disconnection: one diagram and two classes
        self.assertEqual(3, len(list(self.element_factory.select())))



"""
With `get_icon_name` you can retrieve an icon name
for a UML model element.
"""

from gaphor import UML
import re
from functools import singledispatch


TO_KEBAB = re.compile(r"([a-z])([A-Z]+)")


def to_kebab_case(s):
    return TO_KEBAB.sub("\\1-\\2", s).lower()


@singledispatch
def get_icon_name(element):
    """
    Get an icon name for a UML model element.
    """
    return f"gaphor-{to_kebab_case(element.__class__.__name__)}-symbolic"


@get_icon_name.register(UML.Class)
def get_name_for_class(element):
    if isinstance(element, UML.Stereotype):
        return "gaphor-stereotype-symbolic"
    elif element.extension:
        return "gaphor-metaclass-symbolic"
    else:
        return "gaphor-class-symbolic"


@get_icon_name.register(UML.Property)
def get_name_for_property(element):
    if element.association:
        return "gaphor-association-symbolic"
    else:
        return "gaphor-property-symbolic"



""" Tests for :module:`gaphor.misc.generic.event`."""

from __future__ import annotations

from typing import Callable, List
from gaphor.misc.generic.event import Manager


def make_handler(effect: object) -> Callable[[Event], None]:
    return lambda e: e.effects.append(effect)


def create_manager():
    return Manager()


def test_subscribe_single_event():
    events = create_manager()
    events.subscribe(make_handler("handler1"), EventA)
    e = EventA()
    events.handle(e)
    assert len(e.effects) == 1
    assert "handler1" in e.effects


def test_subscribe_via_decorator():
    events = create_manager()
    events.subscriber(EventA)(make_handler("handler1"))
    e = EventA()
    events.handle(e)
    assert len(e.effects) == 1
    assert "handler1" in e.effects


def test_subscribe_event_inheritance():
    events = create_manager()
    events.subscribe(make_handler("handler1"), EventA)
    events.subscribe(make_handler("handler2"), EventB)

    ea = EventA()
    events.handle(ea)
    assert len(ea.effects) == 1
    assert "handler1" in ea.effects

    eb = EventB()
    events.handle(eb)
    assert len(eb.effects) == 2
    assert "handler1" in eb.effects
    assert "handler2" in eb.effects


def test_subscribe_event_multiple_inheritance():
    events = create_manager()
    events.subscribe(make_handler("handler1"), EventA)
    events.subscribe(make_handler("handler2"), EventC)
    events.subscribe(make_handler("handler3"), EventD)

    ea = EventA()
    events.handle(ea)
    assert len(ea.effects) == 1
    assert "handler1" in ea.effects

    ec = EventC()
    events.handle(ec)
    assert len(ec.effects) == 1
    assert "handler2" in ec.effects

    ed = EventD()
    events.handle(ed)
    assert len(ed.effects) == 3
    assert "handler1" in ed.effects
    assert "handler2" in ed.effects
    assert "handler3" in ed.effects


def test_subscribe_no_events():
    events = create_manager()

    ea = EventA()
    events.handle(ea)
    assert len(ea.effects) == 0


def test_subscribe_base_event():
    events = create_manager()
    events.subscribe(make_handler("handler1"), EventA)

    ea = EventB()
    events.handle(ea)
    assert len(ea.effects) == 1
    assert "handler1" in ea.effects


def test_subscribe_event_malformed_multiple_inheritance():
    events = create_manager()
    events.subscribe(make_handler("handler1"), EventA)
    events.subscribe(make_handler("handler2"), EventD)
    events.subscribe(make_handler("handler3"), EventE)

    ea = EventA()
    events.handle(ea)
    assert len(ea.effects) == 1
    assert "handler1" in ea.effects

    ed = EventD()
    events.handle(ed)
    assert len(ed.effects) == 2
    assert "handler1" in ed.effects
    assert "handler2" in ed.effects

    ee = EventE()
    events.handle(ee)
    assert len(ee.effects) == 3
    assert "handler1" in ee.effects
    assert "handler2" in ee.effects
    assert "handler3" in ee.effects


def test_subscribe_event_with_no_subscribers_in_the_middle_of_mro():
    events = create_manager()
    events.subscribe(make_handler("handler1"), Event)
    events.subscribe(make_handler("handler2"), EventB)

    eb = EventB()
    events.handle(eb)
    assert len(eb.effects) == 2
    assert "handler1" in eb.effects
    assert "handler2" in eb.effects


def test_unsubscribe_single_event():
    events = create_manager()
    handler = make_handler("handler1")
    events.subscribe(handler, EventA)
    events.unsubscribe(handler, EventA)
    e = EventA()
    events.handle(e)
    assert len(e.effects) == 0


def test_unsubscribe_event_inheritance():
    events = create_manager()
    handler1 = make_handler("handler1")
    handler2 = make_handler("handler2")
    events.subscribe(handler1, EventA)
    events.subscribe(handler2, EventB)
    events.unsubscribe(handler1, EventA)

    ea = EventA()
    events.handle(ea)
    assert len(ea.effects) == 0

    eb = EventB()
    events.handle(eb)
    assert len(eb.effects) == 1
    assert "handler2" in eb.effects


class Event:
    def __init__(self) -> None:
        self.effects: List[object] = []


class EventA(Event):
    pass


class EventB(EventA):
    pass


class EventC(Event):
    pass


class EventD(EventA, EventC):
    pass


class EventE(EventD, EventA):
    pass



#!/usr/bin/env python

import logging
import os

from gi.repository import Gtk, Gdk


from gaphor.action import action
from gaphor.misc import get_config_dir
from gaphor.abc import ActionProvider
from gaphor.ui.abc import UIComponent
from gaphor.plugins.console.console import GTKInterpreterConsole

log = logging.getLogger(__name__)


class ConsoleWindow(UIComponent, ActionProvider):

    title = "Gaphor Console"
    size = (400, 400)

    def __init__(self, component_registry, main_window, tools_menu):
        self.component_registry = component_registry
        self.main_window = main_window
        tools_menu.add_actions(self)
        self.window = None

    def load_console_py(self, console):
        """Load default script for console. Saves some repetitive typing."""

        console_py = os.path.join(get_config_dir(), "console.py")
        try:
            with open(console_py) as f:
                for line in f:
                    console.push(line)
        except OSError:
            log.info(f"No initiation script {console_py}")

    @action(name="console-window-open", label="_Console")
    def open_console(self):
        if not self.window:
            self.open()
        else:
            self.window.set_property("has-focus", True)

    def open(self):
        console = self.construct()
        self.load_console_py(console)

    def close(self, widget=None):
        if self.window:
            self.window.destroy()
            self.window = None

    def construct(self):
        window = Gtk.Window.new(Gtk.WindowType.TOPLEVEL)
        window.set_transient_for(self.main_window.window)
        window.set_title(self.title)

        console = GTKInterpreterConsole(
            locals={"service": self.component_registry.get_service}
        )
        console.show()
        window.add(console)
        window.show()

        self.window = window

        def key_event(widget, event):
            if (
                event.keyval == Gdk.KEY_d
                and event.get_state() & Gdk.ModifierType.CONTROL_MASK
            ):
                window.destroy()
            return False

        window.connect("key_press_event", key_event)

        window.connect("destroy", self.close)

        return console



"""
Formatting of UML elements like attributes, operations, stereotypes, etc.
"""

import io
import re

from functools import singledispatch

from gaphor.UML import uml2 as UML


@singledispatch
def format(el):
    """
    Format an UML element.
    """
    raise NotImplementedError(
        "Format routine for type %s not implemented yet" % type(el)
    )


@format.register(UML.Property)
def format_property(el, *args, **kwargs):
    """
    Format property or an association end.
    """
    if el.association and not (args or kwargs):
        return format_association_end(el)
    else:
        return format_attribute(el, *args, **kwargs)


def compile(regex):
    return re.compile(regex, re.MULTILINE | re.S)


# Do not render if the name still contains a visibility element
no_render_pat = compile(r"^\s*[+#-]")
vis_map = {"public": "+", "protected": "#", "package": "~", "private": "-"}


def format_attribute(
    el,
    visibility=False,
    is_derived=False,
    type=False,
    multiplicity=False,
    default=False,
    tags=False,
):
    """
    Create a OCL representation of the attribute,
    Returns the attribute as a string.
    If one or more of the parameters (visibility, is_derived, type,
    multiplicity, default and/or tags) is set, only that field is rendered.
    Note that the name of the attribute is always rendered, so a parseable
    string is returned.

    Note that, when some of those parameters are set, parsing the string
    will not give you the same result.
    """
    name = el.name
    if not name:
        name = ""

    if no_render_pat.match(name):
        name = ""

    # Render all fields if they all are set to False
    if not (visibility or is_derived or type or multiplicity or default):
        visibility = is_derived = type = multiplicity = default = True

    s = io.StringIO()

    if visibility:
        s.write(vis_map[el.visibility])
        s.write(" ")

    if is_derived:
        if el.isDerived:
            s.write("/")

    s.write(name)

    if type and el.typeValue:
        s.write(f": {el.typeValue}")

    if multiplicity and el.upperValue:
        if el.lowerValue:
            s.write(f"[{el.lowerValue}..{el.upperValue}]")
        else:
            s.write(f"[{el.upperValue}]")

    if default and el.defaultValue:
        s.write(f" = {el.defaultValue}")

    if tags:
        slots = []
        for slot in el.appliedStereotype[:].slot:
            if slot:
                slots.append(f"{slot.definingFeature.name}={slot.value}")
        if slots:
            s.write(" { %s }" % ", ".join(slots))
    s.seek(0)
    return s.read()


def format_association_end(el):
    """
    Format association end.
    """
    name = ""
    n = io.StringIO()
    if el.name:
        n.write(vis_map[el.visibility])
        n.write(" ")
        if el.isDerived:
            n.write("/")
        if el.name:
            n.write(el.name)
        n.seek(0)
        name = n.read()

    m = io.StringIO()
    if el.upperValue:
        if el.lowerValue:
            m.write(f"{el.lowerValue}..{el.upperValue}")
        else:
            m.write(f"{el.upperValue}")

    slots = []
    for slot in el.appliedStereotype[:].slot:
        if slot:
            slots.append(f"{slot.definingFeature.name}={slot.value}")
    if slots:
        m.write(" { %s }" % ",\n".join(slots))
    m.seek(0)
    mult = m.read()

    return name, mult


@format.register(UML.Operation)
def format_operation(
    el,
    pattern=None,
    visibility=False,
    type=False,
    multiplicity=False,
    default=False,
    tags=False,
    direction=False,
):
    """
    Create a OCL representation of the operation,
    Returns the operation as a string.
    """
    name = el.name
    if not name:
        return ""
    if no_render_pat.match(name):
        return name

    # Render all fields if they all are set to False
    if not (visibility or type or multiplicity or default or tags or direction):
        visibility = type = multiplicity = default = tags = direction = True

    s = io.StringIO()
    if visibility:
        s.write(vis_map[el.visibility])
        s.write(" ")

    s.write(name)
    s.write("(")

    for p in el.formalParameter:
        if direction:
            s.write(p.direction)
            s.write(" ")
        s.write(p.name)
        if type and p.typeValue:
            s.write(f": {p.typeValue}")
        if multiplicity and p.upperValue:
            if p.lowerValue:
                s.write(f"[{p.lowerValue}..{p.upperValue}]")
            else:
                s.write(f"[{p.upperValue}]")
        if default and p.defaultValue:
            s.write(f" = {p.defaultValue}")
        # if p.taggedValue:
        #     tvs = ', '.join(filter(None, map(getattr, p.taggedValue,
        #                                      ['value'] * len(p.taggedValue))))
        #     s.write(' { %s }' % tvs)
        if p is not el.formalParameter[-1]:
            s.write(", ")

    s.write(")")

    rr = el.returnResult and el.returnResult[0]
    if rr:
        if type and rr.typeValue:
            s.write(f": {rr.typeValue}")
        if multiplicity and rr.upperValue:
            if rr.lowerValue:
                s.write(f"[{rr.lowerValue}..{rr.upperValue}]")
            else:
                s.write(f"[{rr.upperValue}]")
        # if rr.taggedValue:
        #    tvs = ', '.join(filter(None, map(getattr, rr.taggedValue,
        #                                     ['value'] * len(rr.taggedValue))))
        #    if tvs:
        #        s.write(' { %s }' % tvs)
    s.seek(0)
    return s.read()


@format.register(UML.Slot)
def format_slot(el):
    return f'{el.definingFeature.name} = "{el.value}"'


@format.register(UML.NamedElement)
def format_namedelement(el):
    """
    Format named element.
    """
    return el.name



#!/usr/bin/env python

"""Gaphor is the simple modeling tool written in Python.

This module allows Gaphor to be launched from the command line.
The main() function sets up the command-line options and arguments and
passes them to the main Application instance.

"""
# Examples of valid version strings
# __version__ = '1.2.3.dev1'  # Development release 1
# __version__ = '1.2.3a1'     # Alpha Release 1
# __version__ = '1.2.3b1'     # Beta Release 1
# __version__ = '1.2.3rc1'    # RC Release 1
# __version__ = '1.2.3'       # Final Release
# __version__ = '1.2.3.post1' # Post Release 1

__version__ = "1.1.1"
__all__ = ["main"]

from optparse import OptionParser
import logging

from gaphor.application import Application

LOG_FORMAT = "%(name)s %(levelname)s %(message)s"


def main():
    """Start Gaphor from the command line.  This function creates an option
    parser for retrieving arguments and options from the command line.  This
    includes a Gaphor model to load.

    The application is then initialized, passing along the option parser.  This
    provides plugins and services with access to the command line options
    and may add their own."""

    parser = OptionParser()

    parser.add_option("-p", "--profiler", action="store_true", help="Run in profiler")
    parser.add_option(
        "-q",
        "--quiet",
        dest="quiet",
        help="Quiet output",
        default=False,
        action="store_true",
    )
    parser.add_option(
        "-v",
        "--verbose",
        dest="verbose",
        help="Verbose output",
        default=False,
        action="store_true",
    )

    options, args = parser.parse_args()

    if options.verbose:
        logging.basicConfig(level=logging.DEBUG, format=LOG_FORMAT)
    elif options.quiet:
        logging.basicConfig(level=logging.WARNING, format=LOG_FORMAT)
    else:
        logging.basicConfig(level=logging.INFO, format=LOG_FORMAT)

    model = None
    try:
        model = args[0]
    except IndexError:
        pass

    if options.profiler:

        import cProfile
        import pstats

        cProfile.runctx(
            "Application.run(model)", globals(), locals(), filename="gaphor.prof"
        )

        profile_stats = pstats.Stats("gaphor.prof")
        profile_stats.strip_dirs().sort_stats("time").print_stats(50)

    else:
        Application.run(model)



""" Tests for :module:`gaphor.misc.generic.registry`."""

import pytest

from typing import Union
from gaphor.misc.generic.registry import Registry, SimpleAxis, TypeAxis


class DummyA:
    pass


class DummyB(DummyA):
    pass


def test_one_axis_no_specificity():
    registry: Registry[object] = Registry(("foo", SimpleAxis()))
    a = object()
    b = object()
    registry.register(a)
    registry.register(b, "foo")

    assert registry.lookup() == a
    assert registry.lookup("foo") == b
    assert registry.lookup("bar") is None


def test_subtyping_on_axes():
    registry: Registry[str] = Registry(("type", TypeAxis()))

    target1 = "one"
    registry.register(target1, object)

    target2 = "two"
    registry.register(target2, DummyA)

    target3 = "three"
    registry.register(target3, DummyB)

    assert registry.lookup(object()) == target1
    assert registry.lookup(DummyA()) == target2
    assert registry.lookup(DummyB()) == target3


def test_query_subtyping_on_axes():
    registry: Registry[str] = Registry(("type", TypeAxis()))

    target1 = "one"
    registry.register(target1, object)

    target2 = "two"
    registry.register(target2, DummyA)

    target3 = "three"
    registry.register(target3, DummyB)

    target4 = "four"
    registry.register(target4, int)

    assert list(registry.query(object())) == [target1]
    assert list(registry.query(DummyA())) == [target2, target1]
    assert list(registry.query(DummyB())) == [target3, target2, target1]
    assert list(registry.query(3)) == [target4, target1]


def test_two_axes():
    registry: Registry[Union[str, object]] = Registry(
        ("type", TypeAxis()), ("name", SimpleAxis())
    )

    target1 = "one"
    registry.register(target1, object)

    target2 = "two"
    registry.register(target2, DummyA)

    target3 = "three"
    registry.register(target3, DummyA, "foo")

    context1 = object()
    assert registry.lookup(context1) == target1

    context2 = DummyB()
    assert registry.lookup(context2) == target2
    assert registry.lookup(context2, "foo") == target3

    target4 = object()
    registry.register(target4, DummyB)

    assert registry.lookup(context2) == target4
    assert registry.lookup(context2, "foo") == target3


def test_get_registration():
    registry: Registry[str] = Registry(("type", TypeAxis()), ("name", SimpleAxis()))
    registry.register("one", object)
    registry.register("two", DummyA, "foo")
    assert registry.get_registration(object) == "one"
    assert registry.get_registration(DummyA, "foo") == "two"
    assert registry.get_registration(object, "foo") is None
    assert registry.get_registration(DummyA) is None


def test_register_too_many_keys():
    registry: Registry[type] = Registry(("name", SimpleAxis()))
    with pytest.raises(ValueError):
        registry.register(object, "one", "two")


def test_lookup_too_many_keys():
    registry: Registry[object] = Registry(("name", SimpleAxis()))
    with pytest.raises(ValueError):
        registry.register(registry.lookup("one", "two"))


def test_conflict_error():
    registry: Registry[Union[object, type]] = Registry(("name", SimpleAxis()))
    registry.register(object(), name="foo")
    with pytest.raises(ValueError):
        registry.register(object, "foo")


def test_skip_nodes():
    registry: Registry[str] = Registry(
        ("one", SimpleAxis()), ("two", SimpleAxis()), ("three", SimpleAxis())
    )
    registry.register("foo", one=1, three=3)
    assert registry.lookup(1, three=3) == "foo"


def test_miss():
    registry: Registry[str] = Registry(
        ("one", SimpleAxis()), ("two", SimpleAxis()), ("three", SimpleAxis())
    )
    registry.register("foo", 1, 2)
    assert registry.lookup(one=1, three=3) is None


def test_bad_lookup():
    registry: Registry[int] = Registry(("name", SimpleAxis()), ("grade", SimpleAxis()))
    with pytest.raises(ValueError):
        registry.register(1, foo=1)
    with pytest.raises(ValueError):
        registry.lookup(foo=1)
    with pytest.raises(ValueError):
        registry.register(1, "foo", name="foo")



from gaphor.diagram.general.comment import CommentItem
from gaphor.diagram.general.commentline import CommentLineItem
from gaphor.diagram.general.simpleitem import Box, Ellipse, Line


def _load():
    from gaphor.diagram.general import connectors, generaleditors, generalpropertypages


_load()



"""
Adapters for the Property Editor.

To register property pages implemented in this module, it is imported in
gaphor.adapter package.

# TODO: make all labels align top-left
# Add hidden columns for list stores where i can put the actual object
# being edited.

TODO:
 - stereotypes
 - association / association ends.
 - Follow HIG guidelines:
   * Leave a 12-pixel border between the edge of the window and
     the nearest controls.
   * Leave a 12-pixel horizontal gap between a control and its label. (The gap
     may be bigger for other controls in the same group, due to differences in
     the lengths of the labels.)
   * Labels must be concise and make sense when taken out of context.
     Otherwise, users relying on screenreaders or similar assistive
     technologies will not always be able to immediately understand the
     relationship between a control and those surrounding it.
   * Assign access keys to all editable controls. Ensure that using the access
     key focuses its associated control.
"""

from typing import Dict, List, Tuple, Type

import abc
from gi.repository import GObject, Gdk, Gtk

import gaphas.item
from gaphas.segment import Segment

from gaphor import UML
from gaphor.core import _, transactional


class _PropertyPages:
    """
    Generic handler for property pages.

    Property pages are collected on type.
    """

    def __init__(self) -> None:
        self.pages: List[Tuple[Type[UML.Element], object]] = []

    def register(self, subject_type):
        def reg(func):
            self.pages.append((subject_type, func))
            return func

        return reg

    def __call__(self, subject):
        for subject_type, func in self.pages:
            if isinstance(subject, subject_type):
                yield func(subject)


PropertyPages = _PropertyPages()


class PropertyPageBase(metaclass=abc.ABCMeta):
    """
    A property page which can display itself in a notebook
    """

    order = 0  # Order number, used for ordered display
    name = "Properties"

    @abc.abstractmethod
    def construct(self):
        """
        Create the page (Gtk.Widget) that belongs to the Property page.

        Returns the page's toplevel widget (Gtk.Widget).
        """


class EditableTreeModel(Gtk.ListStore):
    """Editable GTK tree model based on ListStore model.

    Every row is represented by a list of editable values. Last column
    contains an object, which is being edited (this column is not
    displayed). When editable value in first column is set to empty string
    then object is deleted.

    Last row is empty and contains no object to edit. It allows to enter
    new values.

    When model is edited, then item is requested to be updated on canvas.

    Attributes:
    - _item: diagram item owning tree model
    """

    def __init__(self, item, cols=None):
        """Create new model.

        Parameters:
        - _item: diagram item owning tree model
        - cols: model columns, defaults to [str, object]
        """

        if cols is None:
            cols = (str, object)
        super().__init__(*cols)
        self._item = item

        for data in self._get_rows():
            self.append(data)
        self._add_empty()

    def refresh(self, obj):
        for row in self:
            if row[-1] is obj:
                self._set_object_value(row, len(row) - 1, obj)
                self.row_changed(row.path, row.iter)
                return

    def _get_rows(self):
        """Return rows to be edited.

        Last column has to contain object being edited.
        """

        raise NotImplemented

    def _create_object(self):
        """
        Create new object.
        """
        raise NotImplemented

    def _set_object_value(self, row, col, value):
        """
        Update row's column with a value.
        """
        raise NotImplemented

    def _swap_objects(self, o1, o2):
        """
        Swap two objects. If objects are swapped, then return ``True``.
        """
        raise NotImplemented

    def _get_object(self, iter):
        """
        Get object from ``iter``.
        """
        path = self.get_path(iter)
        return self[path][-1]

    def swap(self, a, b):
        """
        Swap two list rows.
        Parameters:
        - a: path to first row
        - b: path to second row
        """
        if not a or not b:
            return
        o1 = self[a][-1]
        o2 = self[b][-1]
        if o1 and o2 and self._swap_objects(o1, o2):
            # self._item.request_update(matrix=False)
            super().swap(a, b)

    def _add_empty(self):
        """
        Add empty row to the end of the model.
        """
        self.append([None] * self.get_n_columns())

    @transactional
    def set_value(self, iter, col, value):
        row = self[iter][:]

        if col == 0 and not value and row[-1]:
            # kill row and delete object if text of first column is empty
            self.remove(iter)

        elif col == 0 and value and not row[-1]:
            # create new object
            obj = self._create_object()
            row[-1] = obj
            self._set_object_value(row, col, value)
            self._add_empty()

        elif row[-1]:
            self._set_object_value(row, col, value)

        self.set(iter, list(range(len(row))), row)

        self.set(iter, list(range(len(row))), row)

    def remove(self, iter):
        """
        Remove object from GTK model and destroy it.
        """
        obj = self._get_object(iter)
        if obj:
            obj.unlink()
            # self._item.request_update(matrix=False)
            return super().remove(iter)
        else:
            return iter


@transactional
def remove_on_keypress(tree, event):
    """Remove selected items from GTK model on ``backspace`` keypress."""

    k = Gdk.keyval_name(event.keyval).lower()
    if k == "backspace" or k == "kp_delete":
        model, iter = tree.get_selection().get_selected()
        if iter:
            model.remove(iter)


@transactional
def swap_on_keypress(tree, event):
    """Swap selected and previous (or next) items."""

    k = Gdk.keyval_name(event.keyval).lower()
    if k == "equal" or k == "kp_add":
        model, iter = tree.get_selection().get_selected()
        model.swap(iter, model.iter_next(iter))
        return True
    elif k == "minus":
        model, iter = tree.get_selection().get_selected()
        model.swap(iter, model.iter_previous(iter))
        return True


@transactional
def on_text_cell_edited(renderer, path, value, model, col=0):
    """Update editable tree model based on fresh user input."""

    iter = model.get_iter(path)
    model.set_value(iter, col, value)


@transactional
def on_bool_cell_edited(renderer, path, model, col):
    """Update editable tree model based on fresh user input."""

    iter = model.get_iter(path)
    model.set_value(iter, col, renderer.get_active())


class UMLComboModel(Gtk.ListStore):
    """UML combo box model.

    Model allows to easily create a combo box with values and their labels,
    for example

        label1  ->  value1
        label2  ->  value2
        label3  ->  value3

    Labels are displayed by combo box and programmer has easy access to
    values associated with given label.

    Attributes:

    - _data: model data
    - _indices: dictionary of values' indices
    """

    def __init__(self, data):
        super().__init__(str)

        self._indices: Dict[Tuple[str, str], int] = {}
        self._data = data

        # add labels to underlying model and store index information
        for i, (label, value) in enumerate(data):
            self.append([label])
            self._indices[value] = i

    def get_index(self, value):
        """
        Return index of a ``value``.
        """
        return self._indices[value]

    def get_value(self, index):
        """
        Get value for given ``index``.
        """
        return self._data[index][1]


def create_uml_combo(data, callback):
    """
    Create a combo box using ``UMLComboModel`` model.

    Combo box is returned.
    """
    model = UMLComboModel(data)
    combo = Gtk.ComboBox(model=model)
    cell = Gtk.CellRendererText()
    combo.pack_start(cell, True)
    combo.add_attribute(cell, "text", 0)
    combo.connect("changed", callback)
    return combo


def create_hbox_label(adapter, page, label):
    """
    Create a HBox with a label for given property page adapter and page
    itself.
    """
    hbox = Gtk.HBox(spacing=12)
    label = Gtk.Label(label=label)
    # label.set_alignment(0.0, 0.5)
    adapter.size_group.add_widget(label)
    hbox.pack_start(label, False, True, 0)
    page.pack_start(hbox, False, True, 0)
    return hbox


def create_tree_view(model, names, tip="", ro_cols=None):
    """
    Create a tree view for an editable tree model.

    :Parameters:
     model
        Model, for which tree view is created.
     names
        Names of columns.
     tip
        User interface tool tip for tree view.
     ro_cols
        Collection of indices pointing read only columns.
    """
    if ro_cols is None:
        ro_cols = set()

    tree_view = Gtk.TreeView(model=model)

    n = model.get_n_columns() - 1
    for name, i in zip(names, list(range(n))):
        col_type = model.get_column_type(i)
        if col_type == GObject.TYPE_STRING:
            renderer = Gtk.CellRendererText()
            renderer.set_property("editable", i not in ro_cols)
            renderer.set_property("is-expanded", True)
            renderer.connect("edited", on_text_cell_edited, model, i)
            col = Gtk.TreeViewColumn(name, renderer, text=i)
            col.set_expand(True)
            tree_view.append_column(col)
        elif col_type == GObject.TYPE_BOOLEAN:
            renderer = Gtk.CellRendererToggle()
            renderer.set_property("activatable", i not in ro_cols)
            renderer.connect("toggled", on_bool_cell_edited, model, i)
            col = Gtk.TreeViewColumn(name, renderer, active=i)
            col.set_expand(False)
            tree_view.append_column(col)

    tree_view.connect("key_press_event", remove_on_keypress)
    tree_view.connect("key_press_event", swap_on_keypress)

    tip = (
        tip
        + """
Press ENTER to edit item, BS/DEL to remove item.
Use -/= to move items up or down.\
    """
    )
    tree_view.set_tooltip_text(tip)

    return tree_view


@PropertyPages.register(UML.NamedElement)
class NamedElementPropertyPage(PropertyPageBase):
    """An adapter which works for any named item view.

    It also sets up a table view which can be extended.
    """

    order = 10

    NAME_LABEL = _("Name")

    def __init__(self, subject: UML.NamedElement):
        assert subject is None or isinstance(subject, UML.NamedElement), "%s" % type(
            subject
        )
        self.subject = subject
        self.watcher = subject and subject.watcher()
        self.size_group = Gtk.SizeGroup.new(Gtk.SizeGroupMode.HORIZONTAL)

    def construct(self):
        page = Gtk.VBox()

        subject = self.subject
        if not subject:
            return page

        hbox = create_hbox_label(self, page, self.NAME_LABEL)
        entry = Gtk.Entry()
        entry.set_text(subject and subject.name or "")
        hbox.pack_start(entry, True, True, 0)
        page.default = entry

        # monitor subject.name attribute
        changed_id = entry.connect("changed", self._on_name_change)

        def handler(event):
            if event.element is subject and event.new_value is not None:
                entry.handler_block(changed_id)
                entry.set_text(event.new_value)
                entry.handler_unblock(changed_id)

        if self.watcher:
            self.watcher.watch("name", handler).subscribe_all()
            entry.connect("destroy", self.watcher.unsubscribe_all)

        return page

    @transactional
    def _on_name_change(self, entry):
        self.subject.name = entry.get_text()


class NamedItemPropertyPage(NamedElementPropertyPage):
    """
    Base class for named _diagram item_ based adapters.
    """

    def __init__(self, item):
        self.item = item
        super().__init__(item.subject)


@PropertyPages.register(gaphas.item.Line)
class LineStylePage(PropertyPageBase):
    """Basic line style properties: color, orthogonal, etc."""

    order = 400
    name = "Style"

    def __init__(self, item):
        super().__init__()
        self.item = item
        self.size_group = Gtk.SizeGroup(mode=Gtk.SizeGroupMode.HORIZONTAL)

    def construct(self):
        page = Gtk.VBox()

        hbox = Gtk.HBox()
        label = Gtk.Label(label="")
        label.set_justify(Gtk.Justification.LEFT)
        self.size_group.add_widget(label)
        hbox.pack_start(label, False, True, 0)

        button = Gtk.CheckButton(label=_("Orthogonal"))
        button.set_active(self.item.orthogonal)
        button.connect("toggled", self._on_orthogonal_change)
        hbox.pack_start(button, True, True, 0)

        page.pack_start(hbox, False, True, 0)

        hbox = Gtk.HBox()
        label = Gtk.Label(label="")
        label.set_justify(Gtk.Justification.LEFT)
        self.size_group.add_widget(label)
        hbox.pack_start(label, False, True, 0)

        button = Gtk.CheckButton(label=_("Horizontal"))
        button.set_active(self.item.horizontal)
        button.connect("toggled", self._on_horizontal_change)
        hbox.pack_start(button, True, True, 0)

        page.pack_start(hbox, False, True, 0)

        return page

    @transactional
    def _on_orthogonal_change(self, button):
        if len(self.item.handles()) < 3:
            line_segment = Segment(self.item, None)
            line_segment.split_segment(0)
        self.item.orthogonal = button.get_active()
        self.item.canvas.update_now()

    @transactional
    def _on_horizontal_change(self, button):
        self.item.horizontal = button.get_active()
        self.item.canvas.update_now()



"""
The diagram package contains items (to be drawn on the diagram), tools
(used for interacting with the diagram) and interfaces (used for adapting the
diagram).
"""

import gi

gi.require_version("PangoCairo", "1.0")

import gaphor.diagram.actions
import gaphor.diagram.classes
import gaphor.diagram.components
import gaphor.diagram.general
import gaphor.diagram.interactions
import gaphor.diagram.profiles
import gaphor.diagram.states
import gaphor.diagram.usecases



"""
This plugin extends Gaphor with XMI export functionality.
"""

import logging

from gaphor.core import _, action
from gaphor.abc import Service, ActionProvider
from gaphor.plugins.xmiexport import exportmodel
from gaphor.ui.filedialog import FileDialog

logger = logging.getLogger(__name__)


class XMIExport(Service, ActionProvider):
    def __init__(self, element_factory, file_manager, export_menu):
        self.element_factory = element_factory
        self.file_manager = file_manager
        export_menu.add_actions(self)

    def shutdown(self):
        pass

    @action(
        name="file-export-xmi",
        label=_("Export to XMI"),
        tooltip=_("Export model to XMI (XML Model Interchange) format"),
    )
    def execute(self):
        filename = self.file_manager.filename
        if filename:
            filename = filename.replace(".gaphor", ".xmi")
        else:
            filename = "model.xmi"

        file_dialog = FileDialog(
            _("Export model to XMI file"), action="save", filename=filename
        )

        filename = file_dialog.selection

        if filename and len(filename) > 0:
            logger.debug(f"Exporting XMI model to: {filename}")
            export = exportmodel.XMIExport(self.element_factory)
            try:
                export.export(filename)
            except Exception as e:
                logger.error(f"Error while saving model to file {filename}: {e}")



from gaphor.diagram.profiles.extension import ExtensionItem


def _load():
    from gaphor.diagram.profiles import (
        extensionconnect,
        metaclasspropertypage,
        stereotypepage,
    )


_load()



from typing import List, Tuple
from math import pi
from gaphas.geometry import Rectangle

from gaphor.diagram.text import (
    text_point_in_box,
    text_size,
    text_draw,
    text_draw_focus_box,
    TextAlign,
    VerticalAlign,
    FontStyle,
    FontWeight,
    TextDecoration,
)


class Padding:  # Enum
    TOP = 0
    RIGHT = 1
    BOTTOM = 2
    LEFT = 3


def draw_border(box, context, bounding_box):
    cr = context.cairo
    d = box.style("border-radius")
    x, y, width, height = bounding_box

    cr.move_to(x, d)
    if d:
        x1 = width + x
        y1 = height + y
        cr.arc(d, d, d, pi, 1.5 * pi)
        cr.line_to(x1 - d, y)
        cr.arc(x1 - d, d, d, 1.5 * pi, y)
        cr.line_to(x1, y1 - d)
        cr.arc(x1 - d, y1 - d, d, 0, 0.5 * pi)
        cr.line_to(d, y1)
        cr.arc(d, y1 - d, d, 0.5 * pi, pi)
    else:
        cr.rectangle(x, y, width, height)

    cr.close_path()
    cr.stroke()


def draw_top_separator(box, context, bounding_box):
    x, y, w, h = bounding_box
    cr = context.cairo
    cr.move_to(x, y)
    cr.line_to(x + w, y)
    cr.stroke()


def draw_highlight(context):
    highlight_color = (0, 0, 1, 0.4)
    cr = context.cairo
    cr.save()
    try:
        if context.dropzone:
            cr.set_source_rgba(*highlight_color)
            cr.set_line_width(cr.get_line_width() * 3.141)
            cr.stroke_preserve()
    finally:
        cr.restore()


class Box:
    """
    A box like shape.

    Style properties:
    - min-height
    - min-width
    - padding: a tuple (top, right, bottom, left)

    """

    def __init__(self, *children, style={}, draw=None):
        self.children = children
        self.sizes: List[Tuple[int, int]] = []
        self.style = {
            "min-width": 0,
            "min-height": 0,
            "padding": (0, 0, 0, 0),
            "vertical-align": VerticalAlign.MIDDLE,
            "border-radius": 0,
            **style,
        }.__getitem__
        self._draw_border = draw

    def __len__(self):
        return len(self.children)

    def __iter__(self):
        return iter(self.children)

    def __getitem__(self, index):
        return self.children[index]

    def size(self, cr):
        style = self.style
        min_width = style("min-width")
        min_height = style("min-height")
        padding = style("padding")
        self.sizes = sizes = [c.size(cr) for c in self.children]
        if sizes:
            widths, heights = list(zip(*sizes))
            return (
                max(
                    min_width,
                    max(widths) + padding[Padding.RIGHT] + padding[Padding.LEFT],
                ),
                max(
                    min_height,
                    sum(heights) + padding[Padding.TOP] + padding[Padding.BOTTOM],
                ),
            )
        else:
            return min_width, min_height

    def draw(self, context, bounding_box):
        padding = self.style("padding")
        valign = self.style("vertical-align")
        height = sum(h for _w, h in self.sizes)
        if self._draw_border:
            self._draw_border(self, context, bounding_box)
        x = bounding_box.x + padding[Padding.LEFT]
        if valign is VerticalAlign.MIDDLE:
            y = (
                bounding_box.y
                + padding[Padding.TOP]
                + (max(height, bounding_box.height - padding[Padding.TOP]) - height) / 2
            )
        elif valign is VerticalAlign.BOTTOM:
            y = bounding_box.y + bounding_box.height - height - padding[Padding.BOTTOM]
        else:
            y = bounding_box.y + padding[Padding.TOP]
        w = bounding_box.width - padding[Padding.RIGHT] - padding[Padding.LEFT]
        for c, (_w, h) in zip(self.children, self.sizes):
            c.draw(context, Rectangle(x, y, w, h))
            y += h


class IconBox:
    """
    A special type of box: the icon element is given the full width/height and
    all other shapes are drawn below the main icon shape.

    Style properties:
    - min-height
    - min-width
    - vertical-spacing: spacing between icon and children
    - padding: a tuple (top, right, bottom, left)
    """

    def __init__(self, icon, *children, style={}):
        self.icon = icon
        self.children = children
        self.sizes: List[Tuple[int, int]] = []
        self.style = {
            "min-width": 0,
            "min-height": 0,
            "vertical-spacing": 4,
            "padding": (0, 0, 0, 0),
            **style,
        }.__getitem__

    def size(self, cr):
        style = self.style
        min_width = style("min-width")
        min_height = style("min-height")
        padding = style("padding")
        self.sizes = [c.size(cr) for c in self.children]
        width, height = self.icon.size(cr)
        return (
            max(min_width, width + padding[Padding.RIGHT] + padding[Padding.LEFT]),
            max(min_height, height + padding[Padding.TOP] + padding[Padding.BOTTOM]),
        )

    def draw(self, context, bounding_box):
        padding = self.style("padding")
        vertical_spacing = self.style("vertical-spacing")
        x = bounding_box.x + padding[Padding.LEFT]
        y = bounding_box.y + padding[Padding.TOP]
        w = bounding_box.width - padding[Padding.RIGHT] - padding[Padding.LEFT]
        h = bounding_box.height - padding[Padding.TOP] - padding[Padding.BOTTOM]
        self.icon.draw(context, Rectangle(x, y, w, h))
        y = y + bounding_box.height + vertical_spacing
        for c, (cw, ch) in zip(self.children, self.sizes):
            mw = max(w, cw)
            c.draw(context, Rectangle(x - (mw - w) / 2, y, mw, ch))
            y += ch


class Text:
    def __init__(self, text=lambda: "", width=lambda: -1, style={}):
        self._text = text if callable(text) else lambda: text
        self.width = width if callable(width) else lambda: width
        self.style = {
            "width": -1,
            "min-width": 30,
            "min-height": 14,
            "font": "sans 10",
            "font-style": FontStyle.NORMAL,
            "font-weight": None,
            "text-decoration": None,
            "text-align": TextAlign.CENTER,
            "vertical-align": VerticalAlign.MIDDLE,
            "padding": (0, 0, 0, 0),
            **style,
        }.__getitem__

    def text(self):
        try:
            return self._text()
        except AttributeError:
            return ""

    def font(self):
        style = self.style
        return {
            "font": style("font"),
            "font-style": style("font-style"),
            "font-weight": style("font-weight"),
            "text-decoration": style("text-decoration"),
        }

    def size(self, cr):
        min_w = self.style("min-width")
        min_h = self.style("min-height")
        padding = self.style("padding")

        width, height = text_size(cr, self.text(), self.font(), self.width())
        return (
            max(min_w, width + padding[Padding.RIGHT] + padding[Padding.LEFT]),
            max(min_h, height + padding[Padding.TOP] + padding[Padding.BOTTOM]),
        )

    def draw(self, context, bounding_box):
        cr = context.cairo
        min_w = max(self.style("min-width"), bounding_box.width)
        min_h = max(self.style("min-height"), bounding_box.height)
        text_align = self.style("text-align")
        vertical_align = self.style("vertical-align")
        padding = self.style("padding")

        text_box = Rectangle(
            bounding_box.x + padding[Padding.LEFT],
            bounding_box.y + padding[Padding.TOP],
            bounding_box.width - padding[Padding.RIGHT] - padding[Padding.LEFT],
            bounding_box.height - padding[Padding.TOP] - padding[Padding.BOTTOM],
        )

        x, y, w, h = text_draw(
            cr,
            self.text(),
            self.font(),
            lambda w, h: text_point_in_box(
                text_box, (w, h), text_align, vertical_align
            ),
            width=text_box.width,
            default_size=(min_w, min_h),
        )
        return x, y, w, h


class EditableText(Text):
    def draw(self, context, bounding_box):
        x, y, w, h = super().draw(context, bounding_box)
        text_draw_focus_box(context, x, y, w, h)


def draw_default_head(context):
    """
    Default head drawer: move cursor to the first handle.
    """
    context.cairo.move_to(0, 0)


def draw_default_tail(context):
    """
    Default tail drawer: draw line to the last handle.
    """
    context.cairo.line_to(0, 0)


def draw_arrow_head(context):
    cr = context.cairo
    cr.set_dash((), 0)
    cr.move_to(15, -6)
    cr.line_to(0, 0)
    cr.line_to(15, 6)
    cr.stroke()
    cr.move_to(0, 0)


def draw_arrow_tail(context):
    cr = context.cairo
    cr.line_to(0, 0)
    cr.stroke()
    cr.move_to(15, -6)
    cr.line_to(0, 0)
    cr.line_to(15, 6)



"""Classes related (dependency, implementation) adapter connections."""

from gaphor import UML
from gaphor.diagram.connectors import (
    IConnect,
    UnaryRelationshipConnect,
    RelationshipConnect,
)
from gaphor.diagram.presentation import Named, Classified
from gaphor.diagram.classes.dependency import DependencyItem
from gaphor.diagram.classes.implementation import ImplementationItem
from gaphor.diagram.classes.generalization import GeneralizationItem
from gaphor.diagram.classes.association import AssociationItem


@IConnect.register(Named, DependencyItem)
class DependencyConnect(RelationshipConnect):
    """Connect two Named elements using a Dependency."""

    def allow(self, handle, port):
        line = self.line
        element = self.element

        # Element should be a NamedElement
        if not element.subject or not isinstance(element.subject, UML.NamedElement):
            return False

        return super().allow(handle, port)

    def reconnect(self, handle, port):
        line = self.line
        dep = line.subject
        if dep and handle is line.head:
            for s in dep.supplier:
                del dep.supplier[s]
        elif dep and handle is line.tail:
            for c in dep.client:
                del dep.client[c]
        self.reconnect_relationship(
            handle, line.dependency_type.supplier, line.dependency_type.client
        )

    def connect_subject(self, handle):
        """
        TODO: check for existing relationships (use self.relation())
        """

        line = self.line

        if line.auto_dependency:
            canvas = line.canvas
            opposite = line.opposite(handle)

            if handle is line.head:
                client = self.get_connected(opposite).subject
                supplier = self.element.subject
            else:
                client = self.element.subject
                supplier = self.get_connected(opposite).subject
            line.dependency_type = UML.model.dependency_type(client, supplier)

        relation = self.relationship_or_new(
            line.dependency_type,
            line.dependency_type.supplier,
            line.dependency_type.client,
        )
        line.subject = relation


@IConnect.register(Classified, GeneralizationItem)
class GeneralizationConnect(RelationshipConnect):
    """Connect Classifiers with a Generalization relationship."""

    def reconnect(self, handle, port):
        self.reconnect_relationship(
            handle, UML.Generalization.general, UML.Generalization.specific
        )

    def connect_subject(self, handle):
        relation = self.relationship_or_new(
            UML.Generalization, UML.Generalization.general, UML.Generalization.specific
        )
        self.line.subject = relation


@IConnect.register(Classified, AssociationItem)
class AssociationConnect(UnaryRelationshipConnect):
    """Connect association to classifier."""

    def allow(self, handle, port):
        element = self.element

        # Element should be a Classifier
        if not isinstance(element.subject, UML.Classifier):
            return None

        return super().allow(handle, port)

    def connect_subject(self, handle):
        element = self.element
        line = self.line

        c1 = self.get_connected(line.head)
        c2 = self.get_connected(line.tail)
        if c1 and c2:
            head_type = c1.subject
            tail_type = c2.subject

            # First check if we do not already contain the right subject:
            if line.subject:
                end1 = line.subject.memberEnd[0]
                end2 = line.subject.memberEnd[1]
                if (end1.type is head_type and end2.type is tail_type) or (
                    end2.type is head_type and end1.type is tail_type
                ):
                    return

            # Create new association
            relation = UML.model.create_association(head_type, tail_type)
            relation.package = element.canvas.diagram.namespace

            line.head_end.subject = relation.memberEnd[0]
            line.tail_end.subject = relation.memberEnd[1]

            # Do subject itself last, so event handlers can trigger
            line.subject = relation

    def reconnect(self, handle, port):
        line = self.line
        c = self.get_connected(handle)
        if handle is line.head:
            end = line.tail_end
            oend = line.head_end
        elif handle is line.tail:
            end = line.head_end
            oend = line.tail_end
        else:
            raise ValueError("Incorrect handle passed to adapter")

        nav = oend.subject.navigability

        UML.model.set_navigability(line.subject, end.subject, None)  # clear old data

        oend.subject.type = c.subject
        UML.model.set_navigability(line.subject, oend.subject, nav)

    def disconnect_subject(self, handle):
        """
        Disconnect model element.
        Disconnect property (memberEnd) too, in case of end of life for
        Extension
        """
        opposite = self.line.opposite(handle)
        c1 = self.get_connected(handle)
        c2 = self.get_connected(opposite)
        if c1 and c2:
            old = self.line.subject
            del self.line.subject
            del self.line.head_end.subject
            del self.line.tail_end.subject
            if old and len(old.presentation) == 0:
                for e in list(old.memberEnd):
                    e.unlink()
                old.unlink()


@IConnect.register(Named, ImplementationItem)
class ImplementationConnect(RelationshipConnect):
    """Connect Interface and a BehavioredClassifier using an Implementation."""

    def allow(self, handle, port):
        line = self.line
        element = self.element

        # Element at the head should be an Interface
        if handle is line.head and not isinstance(element.subject, UML.Interface):
            return None

        # Element at the tail should be a BehavioredClassifier
        if handle is line.tail and not isinstance(
            element.subject, UML.BehavioredClassifier
        ):
            return None

        return super().allow(handle, port)

    def reconnect(self, handle, port):
        line = self.line
        impl = line.subject
        if handle is line.head:
            for s in impl.contract:
                del impl.contract[s]
        elif handle is line.tail:
            for c in impl.implementatingClassifier:
                del impl.implementatingClassifier[c]
        self.reconnect_relationship(
            handle,
            UML.Implementation.contract,
            UML.Implementation.implementatingClassifier,
        )

    def connect_subject(self, handle):
        """
        Perform implementation relationship connection.
        """
        relation = self.relationship_or_new(
            UML.Implementation,
            UML.Implementation.contract,
            UML.Implementation.implementatingClassifier,
        )
        self.line.subject = relation



"""
A registry for components (e.g. services) and event handling.
"""

from typing import Set, Tuple
from gaphor.abc import Service
from gaphor.application import ComponentLookupError


class ComponentRegistry(Service):
    """
    The ComponentRegistry provides a home for application wide components.
    """

    def __init__(self):
        self._comp: Set[Tuple[object, str]] = set()

    def shutdown(self):
        pass

    def get_service(self, name):
        """Obtain a service used by Gaphor by name.
        E.g. service("element_factory")
        """
        return self.get(Service, name)

    def register(self, component, name):
        self._comp.add((component, name))

    def unregister(self, component):
        self._comp = {(c, n) for c, n in self._comp if not c is component}

    def get(self, base, name):
        found = {(c, n) for c, n in self._comp if isinstance(c, base) and n == name}
        if len(found) > 1:
            raise ComponentLookupError(
                f"More than one component matches {base}+{name}: {found}"
            )
        if len(found) == 0:
            raise ComponentLookupError(
                f"Component with type {base} and name {name} is not registered"
            )
        return next(iter(found))[0]

    def all(self, base):
        return ((c, n) for c, n in self._comp if isinstance(c, base))
